# Copyright (c) 2021 Allthenticate
"""
This module contains the implementation of iBeacons, allowing users to send advertising packets using the iBeacon
format.
"""

# Native modules
import uuid
from typing import List, Optional, Union

# ABle Modules
from able.advertisement import ABleAdvertisement


def check_list(arr: List[bytes]) -> bool:
    """
    Utility function to make sure an object is a list of bytes
    """
    for elem in arr:
        # Check if elem is bytes
        if not isinstance(elem, bytes):
            return False
        # Check that elem is only 1 byte
        if len(elem) != 1:
            return False

    return len(arr) != 0


def validate_args(
    device_uuid: Union[str, uuid.UUID],
    major: List[bytes],
    minor: List[bytes],
    tx_power: List[bytes],
) -> Optional[bool]:
    """
    Utility function that validates the iBeacon data to make sure they conform to the iBeacon protocol
    """
    # Handle type checking
    if not isinstance(device_uuid, (str, uuid.UUID)):
        raise TypeError(f"Expected {device_uuid} to be of type string or UUID")

    if not check_list(major):
        raise TypeError(
            f"Expected {major} to be of a list of bytes (8-bits per byte element)"
        )

    if not check_list(minor):
        raise TypeError(
            f"Expected {major} to be of a list of bytes (8-bits per byte element)"
        )

    if not check_list(tx_power):
        raise TypeError(
            f"Expected {major} to be of a list of bytes (8-bits per byte element)"
        )

    # Handle length checking
    try:
        uuid.UUID(str(device_uuid))
    except ValueError as error:
        raise ValueError(f"Expected {device_uuid} to be a 16 byte UUID") from error

    if len(major) != 2:
        raise ValueError(f"Expected {major} to have length of 2 bytes")

    if len(minor) != 2:
        raise ValueError(f"Expected {minor} to have length of 2 bytes")

    if len(tx_power) != 1:
        raise ValueError(f"Expected {tx_power} to have length of 1 byte")

    return True


class ABleiBeacon(ABleAdvertisement):
    """
    Platform agnostic iBeacon class that creates an advertisement using Apple's iBeacon format
    Reference: https://en.wikipedia.org/wiki/IBeacon

    :param local_name: The name you wish to include within the advertisement packet as the local name
    :type local_name: str
    :param device_uuid: The uuid that you want to broadcast in the iBeacon
    :type device_uuid: Union[str, uuid.UUID]
    :param major: User-defined value to send in iBeacon data
    :type major: List[bytes]
    :param minor: User-defined value to send in iBeacon data
    :type minor: List[bytes]
    :param tx_power: User-defined value to send in iBeacon data
    :type tx_power: List[bytes]
    """

    def __init__(
        self,
        local_name: str,
        device_uuid: Union[str, uuid.UUID],
        major: List[bytes],
        minor: List[bytes],
        tx_power: List[bytes],
    ):
        """
        Initializes the iBeacon object and adds the data to the advertisement's manufacturing data
        """
        try:
            validate_args(device_uuid, major, minor, tx_power)
        except Exception as e:
            raise ValueError("Invalid parameters for initializing an iBeacon") from e

        super().__init__(local_name=local_name)

        # Constant values necessary for iBeacon
        self.company_id = 0x004C
        self.beacon_type = [b"\x02", b"\x15"]

        # User-defined values
        self.uuid = device_uuid
        self.major = major
        self.minor = minor
        self.tx_power = tx_power

        # Add iBeacon data to manufacturer data
        self._add_to_manufacturer_data()

    def _add_to_manufacturer_data(self) -> None:
        """
        Preps the iBeacon data and adds it to manufacturer data
        """
        # Convert uuid to bytes
        if isinstance(self.uuid, str):
            self.uuid = uuid.UUID(self.uuid)

        uuid_bytes = self.uuid.bytes
        uuid_arr = [uuid_bytes[i : i + 1] for i in range(len(uuid_bytes))]
        manu_data = b"".join(
            self.beacon_type + uuid_arr + self.major + self.minor + self.tx_power
        )

        self.add_manufacturer_data(self.company_id, manu_data)



#    Copyright (c)  2021  Allthenticate
"""
This module contains the implementation of ABleCentral (A BLE Central) which serves as an abstraction layer
exposing a berkeley-socket like interface for communicating with centrals which connected via the peripheral
server.

References:
Berkeley Sockets: https://en.wikipedia.org/wiki/Berkeley_sockets
"""

# Native modules
import asyncio
import logging
from functools import wraps
from typing import Callable, Dict, List

# Able modules
from able import ABleApplication, ABleCharacteristic

# Configure logging
logger = logging.getLogger(__name__)


def is_connected_wrapper(method: Callable) -> Callable:
    """
    Wrapper for async methods in ABleCentral which checks whether or not the central is still connected before
    calling the function.

    :raises ConnectionError: if no longer connected to the central, disconnect callback already called

    :param method: the method to wrap with the chec
    :return: the wrapped method
    """

    @wraps(method)
    async def wrapper(self, *args, **kwargs):
        """Wrapper for the is connected wrapper"""
        if not self.is_connected:
            raise ConnectionError(f"No longer connected to {self}")
        if self.is_closing:
            raise ConnectionError(f"Serverside disconnect already initiated to {self}")
        return await method(self, *args, **kwargs)

    return wrapper


class ABleCentral:
    """
    A platform agnostic BLE central class. This wrapper surrounds methods within the peripheral server
    using a berkeley-socket-like API.

    Class attribute `_app` will be the `ABleApplication` that created this central and is required
    to interact with the central itself.

    :param application: the application that created this central
    :param identifier: the identifier used by the application which uniquely identifies this central
    :ivar str identifier: the identifier of the central, this is the mac address on Ubuntu and a uuid on MacOS
    :ivar asyncio.Event disconnected_event: an event that will only be set when the we get a callback indicating a
        disconnect from the hardware
    :ivar bool is_closing: a boolean indicator if the peripheral has begun to tear down the connection, if this is
        `True` all IO will be blocked
    :ivar int mtu: the maximum transmission unit of the central, if any data is attempted to be sent larger than this
        value an exception will be raised,
    """

    def __init__(
        self,
        application: ABleApplication,
        identifier: str,
        adapter_path=None,
        central_obj=None,
    ):
        """Initializes the Central"""

        # Members for interacting with the app
        self._app = application
        self.identifier = identifier

        # State members
        self.disconnected_event: asyncio.Event = asyncio.Event()
        self.is_closing: bool = False  # todo(bernie) implement
        if adapter_path:
            self.adapter_path: str = adapter_path

        # Core Bluetooth object?
        if central_obj:
            self.obj = central_obj

        # Track our data
        self.char_queues: Dict[int, asyncio.Queue[bytes]] = {}

        # Track the char UUIDs which this central is subscribed to
        self.char_subscriptions: List[str] = []

        # WIP
        self.mtu = 256

    @property
    def is_connected(self):
        """Is connected is the same as returning if the disconnected event has not yet been set"""
        return not self.disconnected_event.is_set()

    def __del__(self):
        """TODO(Bernie) is there cleanup we need to do here?"""

    def __str__(self):
        return f"ABleCentral {self.identifier}"

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return hash(self.identifier)

    def __eq__(self, other) -> bool:
        return self.identifier == other.identifier

    @is_connected_wrapper
    async def is_alive(
        self,
        message: bytes = b"ping",
        response_expected: bool = False,
        expected_value: bytes = None,
    ) -> bool:
        """
        The is alive method will send a ping message to the central and await a response if that is expected. If a
        response is expected, the response will be verified against the expected value

        TODO(Bernie): specify a timeout

        :param message: what message should be used as the is alive check (default value=b'ping')
        :param response_expected: `True` if the central will send a response to be checked, `False` otherwise
        :param expected_value: the expected bytes to receive if a response is expected
        :return: `True`, if the central was able to receive the ping (if applicable sent correct response), `False`
            otherwise
        """
        raise NotImplementedError()

    async def close(self) -> None:
        """
        Calls the disconnect command in the ABleApplication that owns this central. This will also set the is closing
        member to true to indicate that the process of tearing down the connection has begun and no more operations
        should happen.

        :return: None
        """
        if not self.is_connected:
            logger.info(f"No longer connected to {self}, close is a NOOP")
            return

        await self._app.disconnect(self)

    @property
    def dbus_path(self) -> str:
        """
        Property which returns a bluezdbus path using the identifier of the device, not that this should only be used
        on the linux backend.

        # TODO(Bernie): Don't allow use if not on linux, or move to somewhere else

        :return: a dbus object path for this device
        """
        return f"{self.adapter_path}/dev_" + self.identifier.replace(":", "_")

    @is_connected_wrapper
    async def send(
        self, data: bytes, characteristic: ABleCharacteristic = None
    ) -> None:
        """
        Berkeley socket style send function. Takes in data and by default notifies the central through the comms
        characteristic but an alternate characteristic can be provided.

        :raises ValueError: if the data is longer than the MTU of the device

        :param data: the data to send to the characteristic
        :param characteristic: an alternate characteristic to send the data on via notifying the device
        :return: None
        """

        # If the size of payload is greater than the MTU of the central, raise a ValueError
        if len(data) > self.mtu:
            raise ValueError(
                "Size of payload is larger than the MTU of the central device."
            )

        # Get the comms characteristic by default unless an alternate is passed in
        if characteristic is None:
            characteristic = self._app.communication_characteristic  # type: ignore

        # Send using the target characteristic

        await characteristic.notify_central_of_change(self, data=data)  # type: ignore

    @is_connected_wrapper
    async def recv(
        self, characteristic: ABleCharacteristic = None, timeout: int = None
    ) -> bytes:
        """
        Berkeley socket recv function, waits until there is data on the recv queue and returns data once it is
        populated, an optional timeout can be passed in.

        :raises TimeoutError: if the queue is empty after the timout passed in

        :param timeout: an optional timeout, defaults to None
        :param characteristic: the characteristic that you want to receive data from
        :return: returns data from this centrals data queue which stores writes the central did to the communications
            characteristic
        """

        if characteristic is None:
            characteristic = self._app.communication_characteristic  # type: ignore

        # Do some magic so if this central disconnects this will terminate
        async def timeout_on_disconnect():
            await self.disconnected_event.wait()
            raise ConnectionError("No longer connected!")

        # This guy will wait to see if data gets put on the characteristic queue from the central
        async def _recv():
            return await asyncio.wait_for(
                self.char_queues[hash(characteristic)].get(), timeout=timeout
            )

        try:
            done, pending = await asyncio.wait(
                [timeout_on_disconnect(), _recv()], return_when=asyncio.FIRST_COMPLETED
            )

            # After doing the wait we need to gather the timout on disconnect coro and cancel it
            gather = asyncio.gather(*pending)
            gather.cancel()
            try:
                await gather
            except asyncio.CancelledError:
                pass

            # Return the result from the wait
            return done.pop().result()
        except ConnectionError as e:
            raise ConnectionError(
                "Unable to recv from central that disconnected"
            ) from e

    def flush_buffers(self) -> None:
        """
        Clears all of the recv queues for the central, this is useful if a previous communication timed out and you
        require a fresh buffer before starting a new interaction.

        :returns: None
        """
        for char, queue in self.char_queues.items():
            while not queue.empty():
                queue.get_nowait()
            logger.debug(f"Cleared the queue for {self} for characteristic hash {char}")



"""
Globals defined for the bluezdbus backend
"""

from enum import Enum


class AdvertisementType(Enum):
    """Bluez defined advertising types"""

    BROADCAST = "broadcast"
    PERIPHERAL = "peripheral"


class DBusInterface(Enum):
    """Base class for interface types"""


class BluezServiceInterfaces(DBusInterface, Enum):
    """Bluez defined service interfaces"""

    BLUEZ_ORG = "org.bluez"
    ADAPTER_INTERFACE = "org.bluez.Adapter1"
    ADVERTISEMENT_INTERFACE = "org.bluez.LEAdvertisement1"
    GATT_CHARACTERISTIC_INTERFACE = "org.bluez.GattCharacteristic1"
    GATT_SERVICE_INTERFACE = "org.bluez.GattService1"
    GATT_MANAGER_INTERFACE = "org.bluez.GattManager1"
    ADVERTISING_MANAGER_INTERFACE = "org.bluez.LEAdvertisingManager1"
    DEVICE_INTERFACE = "org.bluez.Device1"


class DbusServiceInterfaces(DBusInterface, Enum):
    """Dbus defined service interfaces"""

    OBJECT_MANAGER_INTERFACE = "org.freedesktop.DBus.ObjectManager"
    PROPERTIES_INTERFACE = "org.freedesktop.DBus.Properties"


def get_service_path(name: str, index: int):
    """Fetches/generates the path of an service"""
    return f"/org/bluez/{name}/service{index}"


def get_advertisement_object_path(name: str, index: int):
    """Fetches the path of an advertisement"""
    return f"/org/bluez/{name}/advertisement{index}"


def get_characteristic_path(service_path: str, index: int):
    """Fetches the path of an characteristic"""
    return f"{service_path}/char{str(index).zfill(4)}"



#    Copyright (c)  2021  Allthenticate
"""
This module contains the implementation of ABlePeripheralServer (A BLE Peripheral Server) which serves as a wrapper
around the platform specific implementations of BLE applications with services, characteristics and advertisements
within this library and acts as an abstraction layer following the Berkeley Socket API for handling BLE comms.
"""
import asyncio
import logging

# Native modules
import sys
from typing import Callable, List, Optional, Union

# Able modules
from able import ABleApplication
from able.advertisement import ABleAdvertisement
from able.central import ABleCentral
from able.characteristic import ABleCharacteristic
from able.service import ABleService

# Setup logging
logger = logging.getLogger(name=__name__)


class ABlePeripheralServer:
    """
    A platform agnostic Ble Peripheral Server Class. This class serves as the high level client for
    the ABle library and serves as the main interface into the low level platform specific-backends.

    Class attribute `_app` will be the `AbleApplication` for the platform this server is running on (eg. BluezDbus App).

    Note that this class doesn't track the advertisements, services, nor characteristics that the application
    tracks under the hood.

    :param name: The name you wish to bind to this application on the DBus, defaults to "ABle"
    :param auto_recover: Set to `True` if you want the underlying application to attempt to recover on driver or
        other incidents
    :param auto_configure: Set to `True` if you want ABle to do a best effort attempt to configure permissions
        on your system to allow ABle to work, defaults to `False`
    """

    def __init__(
        self,
        name: str = "ABle",
        auto_recover: bool = True,
        adapter_path: Optional[str] = None,
        auto_configure: bool = False,
        loop=asyncio.get_event_loop(),
    ):
        """Constructor method for the ABlePeripheralServer"""
        self.name = name

        # Create an application for your platform with the name provided
        self._app = ABleApplication(
            name=self.name,
            adapter_path=adapter_path,
            auto_recover=auto_recover,
            auto_configure=auto_configure,
        )

        # Track an event loop
        self.loop = loop

    async def setup_peripheral_server(self) -> None:
        """
        Handle all the setup that the underlying application needs to do within the running event loop. This needs to
        be a separate asynchronous function from the initializer for the class because the `setup()` function for the
        app may be async as well.

        TODO(Bernie): Is there a scenario where we have the `__init__()` call this directly?

        :return: None
        """
        # Do any setup the underlying application needs to do
        await self._app.setup()

    async def add_advertisement(self, advertisement: ABleAdvertisement) -> None:
        """
        Add an advertisement to the peripheral server. Note that the underlying application manages
        checking whether or not you are over the advertisement limit, if the advertisement has already
        been added and more on a platform specific basis.

        TODO(Ori): As we develop the OSX backend there may be duplicated code in the `_app.add_advertisement()`
            implementation that we should extract to here?

        :param advertisement: An instance of `ABleAdvertisement` to add to the application this server is running.
        :return: None
        """
        # Add the advertisement to the app
        logger.debug(f"Adding advertisement {advertisement} to the peripheral server.")
        await self._app.add_advertisement(advertisement=advertisement)

    async def remove_advertisement(self, advertisement: ABleAdvertisement) -> None:
        """
        Removes an advertisement from the underlying application, note that the application will handle
        removal with a bunch of validation.

        TODO(Bernie): Ensure that this works at runtime, we should be able to have the APS remove an advert
            and refresh the application to make it vanish.

        :param advertisement: The advertisement you want to have removed from the server.
        :return: None
        """
        logger.debug(
            f"Removing advertisement {advertisement} from the peripheral server."
        )

        await self._app.remove_advertisement(advertisement=advertisement)  # type: ignore

    def add_service(self, service: ABleService) -> None:
        """
        Adds a BLEService to the peripheral server using the underlying application.

        :param service: the service to add to the application
        :return: None
        """
        logger.debug(f"Adding {service} to the peripheral server.")
        self._app.add_service(service)  # type: ignore

    def add_characteristic(
        self,
        service: ABleService,
        characteristic: Union[List[ABleCharacteristic], ABleCharacteristic],
        is_comms_char: bool = False,
    ) -> None:
        """
        Adds a BLECharacteristic to the peripheral server using the underlying application.

        :param is_comms_char: `True` will set the characteristic as the communications characteristic,
            defaults to `False`
        :param service: the service to add the characteristic too
        :param characteristic: the characteristic to add, this can also be a list of ABleCharacteristics and each
            will be added for batch adding.
        :return: None
        """
        if isinstance(characteristic, ABleCharacteristic):
            characteristic = [characteristic]

        for chrc in characteristic:
            logger.debug(f"Adding {chrc} to {service}...")

            # If there are no marked characteristics, set this one as the comms
            if self._app.marked_characteristic is None:
                self._app.add_characteristic(
                    service=service, characteristic=chrc, is_comms_char=True  # type: ignore
                )
            else:
                self._app.add_characteristic(
                    service=service, characteristic=chrc, is_comms_char=is_comms_char  # type: ignore
                )

    def add_connect_callback(self, callback: Callable) -> None:
        """
        Adds a user-defined callback function to the application to be called after connecting to a central device.
        :param callback: the function to be called after a succesfull connection
        :return: None
        """
        if not callable(callback):
            raise TypeError(
                f"The function {callback} that is trying to be added as a connection callback is not callable"
            )

        logger.debug(
            f"Adding user-defined callback {callback} to be called after connection"
        )
        self._app.add_connect_callback(callback)

    def add_disconnect_callback(self, callback: Callable) -> None:
        """
        Adds a user-defined callback function to the application to be called after disconnecting from a central device.
        :param callback: the function to be called after a succesfull disconnect
        :return: None
        """
        if not callable(callback):
            raise TypeError(
                f"The function {callback} that is trying to be added as a disconnect callback is not callable"
            )

        logger.debug(
            f"Adding user-defined callback {callback} to be called after disconnect"
        )
        self._app.add_disconnect_callback(callback)

    async def listen(self, remove_connected: bool = False) -> None:
        """
        Listen is the equivalent its Berkeley Socket API counterpart, it will handle any binding that is required
        and sets the application to be in a "listening state".

        :param remove_connected: if `True` remove all currently connected centrals, (default is `False`)
        :return: None
        """
        if remove_connected:
            raise NotImplementedError("This feature is not yet implemented")

        await self._app.start_interface()

    async def accept(self) -> ABleCentral:
        """
        Blocking call which polls the new connection queue (NCQ) of the underlying application to see if there are any
        new connections. Will only return a ABleCentral once it is present on the NCQ and the app's connected central
        dict.

        :return: A newly connected ABleCentral object ready to be interacted with
        """

        # Wait until the app signals it has a new connection via an event, re checking that there is a client in the
        #   queue before proceeding
        while len(self._app.new_connection_queue) <= 0:
            await self._app.has_new_central.wait()

        # Pop the new central off of the 'queue'
        new_central = self._app.new_connection_queue.pop()
        logger.debug(new_central)

        # If there are no more centrals in the queue, clear the event
        if len(self._app.new_connection_queue) == 0:
            self._app.has_new_central.clear()

        # Return the new central
        return new_central

    async def broadcast(
        self, data: Union[bytes, str], characteristic: ABleCharacteristic = None
    ) -> None:
        """
        This function will notify all connected centrals on the provided characteristic of new data, for example
        notifying all centrals of the battery level, a state of this peripheral, etc.

        # TODO(Bernie): How should we handle this with possibly different MTU's for different centrals

        :param characteristic: the characteristic to notify all connected centrals on,
            defaults to the comms characteristic
        :param data: the data to notify the centrals on
        :return: None
        """

        # If the characteristic is none, default to the communication characteristic
        if characteristic is None:
            characteristic = self._app.communication_characteristic

        # If is a string is passed in, encode it
        if isinstance(data, str):
            data = data.encode("utf-8")

        logger.info(f"Broadcasting {data!r} on {characteristic}.")

        await characteristic.broadcast(data)  # type: ignore

    async def close(self) -> None:
        """
        Equivalent to the socket API's close, this will cleanup all existing connections that the app is managing
        and trigger all cleanup within the app (eg. removing match rules).

        :return: None
        """
        if sys.platform == "darwin":
            self._app.stop()
        elif sys.platform == "linux":
            for connected_central in list(self._app.connected_centrals.values()):
                await connected_central.close()



#    Copyright (c)  2021  Allthenticate
"""
This __init__.py handles all the magic imports such that the same
"""

import sys

# Platform specific
if sys.platform == "linux":
    from able.bluezdbus.advertisement import BluezLEAdvertisement as LEAdvertisement
    from able.bluezdbus.application import BluezApplication as ABleApplication
    from able.bluezdbus.characteristic import BluezCharacteristic as ABleCharacteristic
    from able.bluezdbus.characteristic import (
        BlueZCharacteristicFlags as CharacteristicFlags,
    )
    from able.bluezdbus.service import BluezService as ABleService
elif sys.platform == "darwin":
    from able.corebluetooth.advertisement import CBLEAdvertisement as LEAdvertisement
    from able.corebluetooth.application import CBApplication as ABleApplication
    from able.corebluetooth.characteristic import CBCharacteristic as ABleCharacteristic
    from able.corebluetooth.characteristic import (
        CBCharacteristicFlags as CharacteristicFlags,
    )
    from able.corebluetooth.service import ABleCBService as ABleService
else:
    print(f"{sys.platform} is not yet supported on able!")



"""
Implementation of a CoreBluetooth BLE advertisement using pyobjc as the interface
"""

# Native libraries
import logging
import uuid
from typing import Union

# Setup logging
logger = logging.getLogger(name=__name__)


class CBLEAdvertisement:
    """
    Simple class wrapper of the CoreBluetooth advertisementData dictionary
    For more information see:
    https://developer.apple.com/documentation/corebluetooth/cbperipheralmanager/1393252-startadvertising

    :param local_name: The name you wish to include within advertisement packets as the local name, defaults to `BLE
        Peripheral`
    :type local_name: str
    """

    def __init__(self, advert_wrapper, local_name, *args, **kwargs):
        """Constructor for the CBLEAdvertisement"""

        # Advertisement values
        self.local_name = local_name if local_name is not None else "BLE Peripheral"
        self.service_uuids = []

        # Wrapper class, this is the able advertisement that owns this advert
        self.parent = advert_wrapper
        logger.warning(
            f"Use of args: {args} and {kwargs} not supported on macos backend"
        )

        self.application = None

    def __str__(self):
        return f"CBLEAdvertisement [{self.local_name}, {self.service_uuids}"

    def __repr__(self):
        return str(self)

    def add_service_uuid(self, service_uuid: Union[str, uuid.UUID]) -> None:
        """
        Adds a service uuid to the advertisement object.

        :param service_uuid: What service uuid should be added to this advertisement
        :type service_uuid: Union[str, uuid.UUID]
        :return: None
        """
        if str(service_uuid) not in self.service_uuids:
            self.service_uuids.append(str(service_uuid))

    def add_service_data(
        self, service_uuid: Union[str, uuid.UUID], service_data: Union[str, bytes]
    ) -> None:
        """Not implemented because Core Bluetooth does not support this functionality. See:
        https://developer.apple.com/documentation/corebluetooth/cbperipheralmanager/1393252-startadvertising"""
        raise NotImplementedError(
            "CoreBluetooth does not support adding service data to advertisements"
        )

    async def update_service_data(
        self, service_uuid: Union[str, uuid.UUID], service_data: Union[str, bytes]
    ) -> None:
        """Not implemented because Core Bluetooth does not support this functionality. See:
        https://developer.apple.com/documentation/corebluetooth/cbperipheralmanager/1393252-startadvertising"""
        raise NotImplementedError(
            "CoreBluetooth does not support adding service data to advertisements"
        )

    # pylint: disable=unused-argument,no-self-use
    def add_manufacturer_data(
        self, manufacturer_id: int, manufacturer_data: Union[str, bytes]
    ):
        """Not implemented because Core Bluetooth does not support this functionality. See:
        https://developer.apple.com/documentation/corebluetooth/cbperipheralmanager/1393252-startadvertising"""
        logger.warning(
            "CoreBluetooth does not support adding manufacturer data to advertisements"
        )
        # raise NotImplementedError(
        #     "CoreBluetooth does not support adding manufacturer data to advertisements"
        # )



# Copyright (c) 2021 Allthenticate
"""
This module contains the implementation of the ABleAdvertisement, the top level le_advertisement
class that relies on platform specific implementations but exposes key parts of the API that
is not platform specific.
"""

# Native modules
import logging
import uuid
from typing import List, Union

# ABle Modules
from able import LEAdvertisement

# Configure logging
logger = logging.getLogger(__name__)


class ABleAdvertisement:
    """
    A platform agnostic BLE le_advertisement class. This wrapper illustrates the platform specific
    le_advertisement implementations and utilizes the implementation of each backends specific le_advertisement
    in the `self.le_advertisement` member.

    Note that validation of inputs for member functions is handled in this class instead of having the validation
    duplicated on each backend, the only exception to this is platform specific restrictions.

    :param local_name: The name you wish to include within le_advertisement packets as the local name
    """

    def __init__(self, local_name: str, *args, **kwargs):
        """Initializes the mixin le_advertisement class"""
        self.local_name = local_name

        # Handle the platform specific le_advertisement object passing in any kwargs
        self.le_advertisement = LEAdvertisement(self, self.local_name, *args, **kwargs)

        logger.info(
            f"Initialized ABleAdvertisement using {type(self.le_advertisement)}"
        )

    def __str__(self):
        return f"ABleAdvertisement using the {type(self.le_advertisement)} advertising backend"

    def __repr__(self):
        return f"ABleAdvertisement({self.local_name})"

    def add_service_uuid(self, service_uuid: Union[str, uuid.UUID]) -> None:
        """
        Adds a service uuid to the le_advertisement. Please see the Bluetooth SIG's specification
        for service uuids and which uuids are assigned/reserved.

        :raises ValueError: if `service_uuid` is not of type str or uuid.UUID

        :param service_uuid: the service uuid to add to the le_advertisement
        :return: None
        """
        # Handle type checking
        if not isinstance(service_uuid, (str, uuid.UUID)):
            raise TypeError(f"Expected {service_uuid!r} to be a str or uuid.UUID")

        return self.le_advertisement.add_service_uuid(service_uuid)

    async def update_service_uuids(
        self, service_uuids: List[Union[str, uuid.UUID]]
    ) -> None:
        """
        Update service uuids of the advertisement

        :param service_uuids: the new service uuids to be advertised (these will replace the old ones)
        """

        # Clear current service_uuids
        self.le_advertisement.service_uuids = []

        # Add each new one
        for service_uuid in service_uuids:
            self.le_advertisement.add_service_uuid(service_uuid)

        await self.le_advertisement.application.refresh_advertisement(self)

    def add_service_data(
        self, service_uuid: Union[str, uuid.UUID], service_data: Union[str, bytes]
    ) -> None:
        """
        Adds service data for a service uuid within the le_advertisement. Note that the service uuid
        must be added before the service data can be added.

        :raises ValueError: if `service_uuid` is not of type str or uuid.UUID or if `service_data` is not
            of type str or bytes

        :param service_uuid: The uuid of the service to add the service data to, this must be a service uuid
            you have already added to the le_advertisement
        :param service_data: The service data to include in the le_advertisement under the service uuid
        :return: None
        """
        # Handle type checking
        if not isinstance(service_uuid, (str, uuid.UUID)):
            raise TypeError(f"Expected {service_uuid!r} to be a str or uuid.UUID")
        if not isinstance(service_data, (str, bytes)):
            raise TypeError(f"Expected {service_data!r} to be a str or bytes")

        return self.le_advertisement.add_service_data(service_uuid, service_data)

    async def update_service_data(
        self, service_uuid: Union[str, uuid.UUID], service_data: Union[str, bytes]
    ) -> None:
        """
        Update previously set service data with new data and refresh the le_advertisement using the bound
        application. This can be only called after the `ABlePeripheralServer` that owns this le_advertisement
        has started advertising.

        :raises ValueError: if `service_uuid` is not of type str or uuid.UUID or if `service_data` is not
            of type str or bytes

        :param service_uuid: the uuid of the service whose service data you are updating.
        :param service_data:  the service data to overwrite the previous service data with.
        :return: None
        """
        # Handle type checking
        if not isinstance(service_uuid, (str, uuid.UUID)):
            raise TypeError(f"Expected {service_uuid!r} to be a str or uuid.UUID")
        if not isinstance(service_data, (str, bytes)):
            raise TypeError(f"Expected {service_data!r} to be a str or bytes")

        return await self.le_advertisement.update_service_data(
            service_uuid, service_data
        )

    def add_manufacturer_data(
        self, manufacturer_id: int, manufacturer_data: Union[str, bytes]
    ) -> None:
        """
        Updates the manufacturer data of the le_advertisement to include the specified manufacturer data. Refer to the
        reference doc above for the limitations on manufacturer data.

        :raises ValueError: if `manufacturer_id` is not of type 16-bit unsigned int or `manufacturer_data` is not
            of type str or bytes

        :param manufacturer_id: the id of the manufacturer to add the manufacturer data to, see the Bluetooth
            Special Interest groups website for what manufacturer id's are valid to be used in your use case.
        :param manufacturer_data: the manufacturer data to add
        :return: None
        """
        # Handle type checking
        if not isinstance(manufacturer_id, int) or not 0 <= manufacturer_id <= 65535:
            raise TypeError(
                f"Expected {manufacturer_id!r} to be an unsigned 16-bit integer!"
            )
        if not isinstance(manufacturer_data, (str, bytes)):
            raise TypeError(f"Expected {manufacturer_data!r} to be a str or bytes")

        return self.le_advertisement.add_manufacturer_data(
            manufacturer_id, manufacturer_data
        )



"""CoreBluetooth service in python"""

# Copyright (c) 2021 Allthenticate


import uuid
from typing import Union

from CoreBluetooth import CBUUID, CBMutableService  # type: ignore

from able.service import ABleService


class ABleCBService(ABleService):
    """
    Wrapper around the CoreBluetooth service (CBMutableService)

    :param service_uuid: the uuid for this service
    :param is_primary: whether or not this is a primary service, secondary services are rarely used as they
        are intended to be used in other services, defaults to True
    """

    def __init__(
        self, service_uuid: Union[uuid.UUID, str], is_primary: bool = True
    ) -> None:
        """Initialize the ABleCBService"""

        super().__init__(service_uuid)

        cb_service_uuid: CBUUID = CBUUID.alloc().initWithString_(str(service_uuid))

        cb_service: CBMutableService = CBMutableService.alloc().initWithType_primary_(
            cb_service_uuid, is_primary
        )

        self.obj = cb_service



#    Copyright (c)  2021  Allthenticate
"""
References for this implementation of a Bluez Characteristic using dbus-next

GATT Characteristic: https://gitlab.com/allthenticate/dependencies/bluez/-/blob/master/doc/gatt-api.txt
BLE Core Specification: https://www.bluetooth.com/specifications/specs/core-specification/
"""

# Native dependencies
import asyncio
import logging
import multiprocessing
import uuid
from enum import Enum
from typing import TYPE_CHECKING, Optional, Union

# Other dependencies
from dbus_next import Variant  # type: ignore
from dbus_next.constants import PropertyAccess  # type: ignore
from dbus_next.service import ServiceInterface, dbus_property, method  # type: ignore

# Able dependencies
from able.bluezdbus import BluezServiceInterfaces
from able.bluezdbus.service import BluezService
from able.bluezdbus.utils import parse_identifier_from_path

# This is a cyclic import
from able.characteristic import ABleCharacteristic

if TYPE_CHECKING:
    from able.bluezdbus.application import BluezApplication
    from able.central import ABleCentral

logger = logging.getLogger(name=__name__)


class BlueZCharacteristicFlags(Enum):
    """Enum for possible characteristic flags and permissions"""

    AUTHENTICATED_SIGNED_WRITES = "authenticated_signed_writes"
    AUTHORIZE = "authorize"
    BROADCAST = "broadcast"
    ENCRYPT_AUTHENTICATED_READ = "encrypt_authenticated_read"
    ENCRYPT_AUTHENTICATED_WRITE = "encrypt_authenticated_write"
    ENCRYPT_READ = "encrypt_read"
    ENCRYPT_WRITE = "encrypt_write"
    EXTENDED_PROPERTIES = "extended_properties"
    INDICATE = "indicate"
    NOTIFY = "notify"
    READ = "read"
    RELIABLE_WRITE = "reliable_write"
    SECURE_READ = "secure_read"
    SECURE_WRITE = "secure_write"
    WRITABLE_AUXILIARIES = "writable_auxiliaries"
    WRITE = "write"
    WRITE_WITHOUT_RESPONSE = "write_without_response"


# noinspection PyPep8Naming
class BluezCharacteristic(ABleCharacteristic, ServiceInterface):
    """
    Class implementation of a BlE GATT Characteristic using Dbus-Next
    """

    def __init__(
        self,
        characteristic_uuid: uuid.UUID,
        value=bytes(0),
        flags=None,
        nickname: str = None,
        notify_on_write: bool = False,
    ):
        super().__init__(
            characteristic_uuid,
            value,
            flags,
            nickname,
            notify_on_write,
            name=BluezServiceInterfaces.GATT_CHARACTERISTIC_INTERFACE.value,
        )

        # DBus Members
        self._path = None

        # Additional Members
        self.application: Optional["BluezApplication"] = None
        self.subscribed_clients: int = 0
        self.subscribed_clients_lock: multiprocessing.synchronize.Lock = (
            multiprocessing.Lock()
        )

        # Keep track of whether or not this characteristic is the char for socket comms
        self.is_marked = False

        logger.debug(f"Setting up characteristic with flags: {self.flags}")

    def __str__(self):
        return (
            f"BluezCharacteristic [{self.uuid}]"
            if self.nickname is None
            else self.nickname
        )

    def __repr__(self):
        return str(self)

    def __setattr__(self, key, value):
        self.__dict__[key] = value

        # If the key is value, we need to emit the properties changed to bluez so it catches the change
        if key == "value":
            self.emit_properties_changed({"Value": self.value})

    @dbus_property(PropertyAccess.READ, name="UUID")
    def _uuid(self) -> "s":  # type: ignore
        """
        Dbus property to return the UUID of this characteristic.

        :return: 128-bit characteristic UUID.
        """
        return str(self.uuid)

    @dbus_property(PropertyAccess.READ, name="Service")
    def _service(self) -> "o":  # type: ignore
        """
        Dbus property to read the object path of the service this characteristic is owned by.

        :return: Object path of the GATT service the characteristic
                        belongs to.
        """
        if not isinstance(self.service, BluezService):
            raise RuntimeError(
                f"Service of characteristic is not of type BluezService, is {self.service!r}"
            )

        return self.service.path

    @dbus_property(PropertyAccess.READ, name="Value")
    def _value(self) -> "ay":  # type: ignore
        """
        Dbus property to read the current value of the characteristic.

        :return: The cached value of the characteristic. This property
                        gets updated only after a successful read request and
                        when a notification or indication is received, upon
                        which a PropertiesChanged signal will be emitted.
        """
        return self.value

    @dbus_property(PropertyAccess.READ, name="Notifying")
    def _notifying(self) -> "b":  # type: ignore
        """
        Dbus property to read whether or not this characteristic is notifying.

        :return: True, if notifications or indications on this
                        characteristic are currently enabled.
        """
        return self.notifying

    @dbus_property(PropertyAccess.READ, name="Flags")
    def _get_flags(self) -> "as":  # type: ignore
        """
        Dbus property to get the flags associated with the characteristic.

        :return: Defines how the characteristic value can be used.
        """
        return self.flags

    @method(name="ReadValue")
    def _get_value(self, options: "a{sv}") -> "ay":  # type: ignore
        """
        Handles the request from bluez to read the value of the characteristic.

        TODO(Bernie): How should we treat the options here, should we update mtu and such?

        :param options: A dictionary of options provided by bluez regarding the read request. Possible
            values are: "offset": uint16 offset, "mtu": Exchanged MTU (Server only),
            "device": Object Device (Server only)
        :return: the value of the characteristic
        """
        logger.debug(f"Read occurred with options: {options} to {self}")
        return self.value

    @method(name="WriteValue")
    def _set_value(self, value: "ay", options: "a{sv}"):  # type: ignore
        """
        Handles the request from bluez over dbus to write a value to the characteristic.

        TODO(Bernie): Evaluate if we should treat writes from centrals which were connected before the
            service started similar to connects

        :param value: The value to update the characteristic to
        :param options: a dictionary of options regarding the write request. Possible values are:
            "offset": Start offset, "type": `str`, "mtu": Exchanged MTU (Server only),
            "device": Device path (Server only), "link": Link type (Server only),
            "prepare-authorize": `True` if prepare
        :return: None
        """
        if self.application is None:
            raise RuntimeError(
                "No application bound for characteristic, unable to perform actions"
            )

        logger.debug(f"Write occurred to {self} [data: [{value[:20]}]")

        identifier: str = parse_identifier_from_path(options.get("device").value)

        # Fetch the central from the app, if not found
        if identifier not in self.application.connected_centrals:
            logger.warning(
                f"Received a write from {identifier} who we aren't tracking, ignoring"
            )
            return

        # Add the data to the recv queue
        writer: "ABleCentral" = self.application.connected_centrals[identifier]
        asyncio.create_task(writer.char_queues[hash(self)].put(value))

        # Update the MTU if it is provided in the options
        mtu = options.get("mtu", Variant("b", False)).value
        if mtu:
            logger.debug(
                f"Updating {writer} mtu to {mtu} after a write to a characteristic"
            )
            writer.mtu = mtu

        if self.notify_on_write:
            logger.debug(
                f"Updating value of {self} after a write and notifying all centrals subscribed"
            )
            self.value = value
        else:
            logger.debug(f"Not notifying after updating value of {self}")
            self.__dict__["value"] = value

    @method(name="StartNotify")
    def _enable_notifications(self):
        """
        Handles request from bluez which indicates that a central wants to be notified on this characteristic.

        :return: None
        """
        logger.debug(f"{self} is now notifying, incrementing the subscribed semaphore.")

        # Always start notifying and mark as ready
        self.client_subscribed.set()

        # Update the semaphore
        # with self.subscribed_clients_lock:
        self.subscribed_clients += 1

    @method(name="StopNotify")
    def _stop_notify(self):
        """
        Handles the request from bluez which indicates that a central no longer wants to be notified on this
        characteristic.

        :return: None
        """
        logger.debug(
            f"A client has unsubscribed to {self}, decrementing the subscribed semaphore."
        )

        # Decrement the semaphore
        # with self.subscribed_clients_lock:
        self.subscribed_clients -= 1

        # If no more clients are subscribed, stop notifying and clear the ready event
        if self.subscribed_clients <= 0:
            logger.debug(f"There no longer any clients subscribed to {self}")
            self.client_subscribed.clear()

    async def notify_central_of_change(  # pylint: disable=arguments-differ, unused-argument
        self, central: "ABleCentral", data: bytes, *args
    ) -> None:
        """
        Makes a bluez call to notify only one central of a change to this characteristic.

        :param central: the central to notify or indicate to with the new data
        :param data: the data to 'send' to the central by setting the value of this characteristic to and
            notifying only that central
        :return: None
        """
        if not self.notifying:
            raise RuntimeWarning(
                "No central has indicated they want notifications on this characteristic, a central "
                "must subscribe before notifying/indicating"
            )

        if self.application is None or self.application.gatt_manager_interface is None:
            raise RuntimeError(
                "Unable to notify central of change without a bound application"
            )

        try:
            if self.application.supports_directed_notifications:
                await self.application.gatt_manager_interface.call_notify_characteristic_changed(
                    self.application.path,
                    central.dbus_path,
                    self.path,
                    data,
                )
            else:
                logger.debug(
                    "This client does not support directed notifications, broadcasting..."
                )
                await self.broadcast(data)
        except Exception:
            logger.exception(
                f"Exception caught when trying to notify {central}, are you sure the central is setup"
                "to notify on the characteristic?"
            )

    async def broadcast(self, data: Union[bytes, str]) -> None:
        """Broadcast characteristic's new value to every subscribed central

        :param data: the data to notify the centrals on
        """
        # Convert to bytes if it is a string
        if isinstance(data, str):
            data = data.encode("utf8")

        self.value = data

    @property
    def path(self) -> str:
        """Path property, raises an exception if not properly defined"""
        if self._path is None or not isinstance(self._path, str):
            raise RuntimeError(
                "Attempting to fetch path of an unset characteristic, has it been added correctly?"
            )
        return self._path



"""
This is a helper file used for configuring the bluez conf file
"""

import logging
import os
import sys
import xml.etree.ElementTree as ET
from typing import Optional

# Setup logging
logger = logging.getLogger(name=__name__)


def configure_bluez_conf_file(
    peripheral_name: str = "ABle", user_name: str = "root", path: Optional[str] = None
) -> bool:
    """
    This function (must have root permissions to modify the needed file) will modify the conf file dbus
    uses for Bluez such that a user can run an application with a given name

    :param peripheral_name: The name of your application (defaults to ABle)
    :param user_name: The user that will be granted permission (defaults to root)
    :return: True if the configuration file was changed, otherwise False
    """
    # Make sure they are running as root
    if os.geteuid() != 0:
        logger.error("This configuration can only be ran as root!")
        sys.exit(1)

    # Parse the current config file
    if path is None:
        path = "/etc/dbus-1/system.d/bluetooth.conf"
        logger.info(f"No path given for updating permissions, defaulting to {path}")

    tree = ET.parse(path)
    config = tree.getroot()

    root_policy = None
    user_policy = None

    did_change = False

    # First find the user="root" policy
    for policy in config:

        # Check for the root policy
        if policy.attrib == {"user": "root"}:
            root_policy = policy
            continue

        # Check for user policy
        if policy.attrib == {"user": user_name}:
            user_policy = policy
            continue

    # Let's validate that the root policy has a rule to send to this application, root needs to be able to send
    if not root_policy.findall(
        f"allow[@send_destination='org.bluez.{peripheral_name}']"
    ):
        logger.debug(f"Send destination for {peripheral_name} not found, inserting...")
        child = ET.Element(
            "allow", {"send_destination": f"org.bluez.{peripheral_name}"}
        )
        root_policy.append(child)
        did_change = True
    else:
        logger.debug("Root policy already configured...")

    # Make sure a user policy exists
    if user_policy is None:
        logger.debug(f"No user policy exists, creating one for {user_name}")
        user_policy = ET.Element("policy", {"user": user_name})
        config.append(user_policy)
        did_change = True
    else:
        logger.debug(f"A user policy already exists for {user_name}")

    # Make sure all the rules in the user policy exist, the user needs to be own the service org.bluez.<peripheral name>
    if not user_policy.findall(f"allow[@own='org.bluez.{peripheral_name}']"):
        logger.debug(
            f"Own rule for {user_name} for application {peripheral_name} not found, inserting..."
        )
        child = ET.Element("allow", {"own": f"org.bluez.{peripheral_name}"})
        user_policy.append(child)
        did_change = True
    else:
        logger.debug("User policy is already configured.")

    tree.write(path)

    return did_change



#  Copyright (c)  2021  Allthenticate
"""
This module contains utility functions for interacting over dbus using dbus-next

For references of this implementation see:

DBUS: https://dbus.freedesktop.org/doc/dbus-specification.html
dbus-next: https://python-dbus-next.readthedocs.io/_/downloads/en/latest/pdf/
"""

# Dependencies
import re

# 3rd party
from typing import Optional

from dbus_next import Message  # type: ignore
from dbus_next.aio import MessageBus  # type: ignore
from dbus_next.constants import MessageType  # type: ignore


async def add_match(bus: MessageBus, match_rule: str = "type='signal'") -> None:
    """
    Calls the function `AddMatch` to add a match rule so when signals that match the match rule are emitted,
    the bus' handler will receive them.

    :param bus: the bus which will serve as the handler
    :type bus: MessageBus
    :param match_rule: the match rule to filter all messages going over the bus (default is type='signal')
    :type match_rule: str
    :return: None
    :rtype: None
    """
    reply: Optional[Message] = await bus.call(
        Message(
            destination="org.freedesktop.DBus",
            path="/org/freedesktop/DBus",
            interface="org.freedesktop.DBus",
            member="AddMatch",
            signature="s",
            body=[match_rule],
        )
    )

    # Validate the response
    if reply is None:
        raise RuntimeWarning("No message back from adding match rule")

    assert reply.message_type == MessageType.METHOD_RETURN, reply.error_name


async def remove_match(bus: MessageBus, match_rule: str = "type='signal'") -> None:
    """
    Calls the function `RemoveMatch` to remove a match rule so when signals that match the match rule are emitted,
    the bus' handler will NO longer receive them.

    :param bus: the bus which was the handler
    :type bus: MessageBus
    :param match_rule: the match rule to filter all messages going over the bus (default is type='signal')
    :type match_rule: str
    :return: None
    :rtype: None
    """
    reply: Optional[Message] = await bus.call(
        Message(
            destination="org.freedesktop.DBus",
            path="/org/freedesktop/DBus",
            interface="org.freedesktop.DBus",
            member="RemoveMatch",
            signature="s",
            body=[match_rule],
        )
    )

    # Validate the response
    if reply is None:
        raise RuntimeWarning("No message back from removing match rule")

    assert reply.message_type == MessageType.METHOD_RETURN, reply.error_name


def parse_identifier_from_path(dbus_path: str) -> str:
    """
    Parses the identifier (eg. mac address) of a device from the dbus path

    ex. /org/bluez/hci0/dev_6B_0A_12_1A_7D_60

    :param dbus_path: the dbus path of the device
    :type dbus_path: str
    :return: the parsed out identifier
    :rtype: str
    """
    regex = re.compile(r"(?:[0-9a-fA-F]_?){12}")
    match = re.findall(regex, dbus_path)

    if not match:
        ValueError("Inputted a dbus path that did not contain a mac address")

    # Return the entry with underscores swapped
    return match[0].replace("_", ":")



"""CoreBluetooth Errors"""

from enum import Enum


class CBATTError(Enum):
    # pylint: disable=invalid-name
    """CBATTError enumeration"""

    Success = 0x0
    InvalidHandle = 0x1
    ReadNotPermitted = 0x2
    WriteNotPermitted = 0x3
    InvalidPdu = 0x4
    InsufficientAuthentication = 0x5
    RequestNotSupported = 0x6
    InvalidOffset = 0x7
    InsufficientAuthorization = 0x8
    PrepareQueueFull = 0x9
    AttributeNotFound = 0xA
    AttributeNotLong = 0xB
    InsufficientEncryptionKeySize = 0xC
    InvalidAttributeValueLength = 0xD
    UnlikelyError = 0xE
    InsufficientEncryption = 0xF
    UnsupportedGroupType = 0x10
    InsufficientResources = 0x11



#    Copyright (c)  2021  Allthenticate
"""
For references of this implementation see of a bluezdbus application:

Advertising: https://gitlab.com/allthenticate/dependencies/bluez/-/blob/master/doc/advertising-api.txt
GATT: https://gitlab.com/allthenticate/dependencies/bluez/-/blob/master/doc/gatt-api.txt
DBUS: https://dbus.freedesktop.org/doc/dbus-specification.html
dbus-next: https://python-dbus-next.readthedocs.io/_/downloads/en/latest/pdf/
"""

# Native libraries
import asyncio
import logging
import re
import subprocess
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Set

# Dependencies
from dbus_next import DBusError, Message, Variant  # type: ignore
from dbus_next.aio import MessageBus, ProxyInterface, ProxyObject  # type: ignore
from dbus_next.constants import BusType  # type: ignore
from dbus_next.service import ServiceInterface  # type: ignore

from able.bluezdbus import (
    BluezServiceInterfaces,
    DBusInterface,
    DbusServiceInterfaces,
    get_advertisement_object_path,
    get_characteristic_path,
    get_service_path,
)
from able.bluezdbus.advertisement import BluezLEAdvertisement
from able.bluezdbus.characteristic import BluezCharacteristic
from able.bluezdbus.configurator import configure_bluez_conf_file
from able.bluezdbus.service import BluezService
from able.bluezdbus.utils import add_match, remove_match

# Able Dependencies
from able.utils import DeviceCallbackData

if TYPE_CHECKING:
    from able.advertisement import ABleAdvertisement
    from able.central import ABleCentral
    from able.characteristic import ABleCharacteristic
    from able.service import ABleService

# Setup logging
logger = logging.getLogger(name=__name__)

BLUETOOTH_FIVE_HCI_VERSION = 0x9


def _check_support_for_always_advertising(adapter_path: str) -> bool:
    """
    Advertising while connected is only supported on adapters with HCI version ^5.0, this function uses
    hciconfig to check the version the adapter specified.

    :param adapter_path: the adapter path to check (eg. /org/bluez/hci0)
    :return: `True` if multiple
    """
    try:
        # Get the HCI version from hciconfig
        output = subprocess.run(
            ["hciconfig", "-a", adapter_path], stdout=subprocess.PIPE, check=True
        )

        # Find the version number using a re
        version_match = re.search(
            r"HCI Version:[^|\n]*\((.*)\)", output.stdout.decode()
        )

        if version_match is None:
            logger.warning("Unable to parse HCI version from hciconfig")
            return False

        version = int(version_match.group(1), 16)
        logger.debug(
            f"Detected HCI version is {version}, needs to be {BLUETOOTH_FIVE_HCI_VERSION} for "
            f"advertising while connected"
        )

        # Return if it is greater than or equal to bluetooth 5
        return version >= BLUETOOTH_FIVE_HCI_VERSION
    except Exception:
        logger.exception(
            "Unable to determine if always advertising is supported, defaulting to False"
        )
        return False


def _check_support_for_directed_notifications(gatt_manager: ProxyInterface) -> bool:
    """
    Uses the proxy interface to determine whether or not a custom patch enabling device specific notifications
    is supported. The method name we are looking for is `call_notify_characteristic_changed`

    :param gatt_manager: The proxy interface of the gatt manager
    :return: `True` if the patched method is present and callable, `False` otherwise
    """
    return "call_notify_characteristic_changed" in gatt_manager.__dict__


class BluezApplication(ServiceInterface):
    """
    The high level application that abstracts Dbus Calls

    :param name: the name that this application should request on the dbus
    :param auto_recover: if `True` will try to recover if bluez restarts, defaults to `False`
    :param adapter_path: what is the path to the hardware adapter, will be the first one discovered if not provided,
        eg. /org/bluez/hci0
    :param auto_configure: if `True` will try to set bluez settings at startup, defaults to `False`
    """

    def __init__(
        self,
        name: str,
        adapter_path: Optional[str] = None,
        auto_recover: bool = False,
        auto_configure: bool = False,
    ):
        """Initializer for the bluezdbus application"""
        super().__init__(BluezServiceInterfaces.GATT_MANAGER_INTERFACE.value)

        # Dbus related
        self.name = name
        self.path = "/"
        self._bus: Optional[MessageBus] = None
        self.gatt_manager_interface = None
        self.adapter_interface: Optional[ProxyInterface] = None
        self.proxy_objects: Dict[str, Dict[str, ProxyObject]] = {}
        self.match_rules: Set[str] = set()
        self.auto_recover = auto_recover
        self.error_state = False
        self.dbus_lock: Optional[asyncio.Lock] = None
        self.supports_directed_notifications: bool = False

        # Advertising related
        self.advertisements: Dict[int, "ABleAdvertisement"] = {}
        self.advertising_manager: Optional[ProxyInterface] = None
        self.adapter_path = adapter_path
        self.adapter_properties: Optional[ProxyInterface] = None
        self.supported_includes = None
        self.advertising_lock: Optional[asyncio.Lock] = None

        # Services related
        self.services: List[BluezService] = []

        # Params
        self.max_advertisements = 5
        self.advertising_capabilities: Dict[str, int] = {}
        self.always_advertising_supported = False

        # Connection related
        self.has_new_central = asyncio.Event()
        self.new_connection_queue: Set["ABleCentral"] = set()
        self.connected_centrals: Dict[str, "ABleCentral"] = {}
        self.user_connect_callback: Optional[Callable] = None
        self.user_disconnect_callback: Optional[Callable] = None

        # Communication related
        self.marked_characteristic: Optional[BluezCharacteristic] = None

        # If set to autoconfigure, try to do so for root
        if auto_configure:
            logger.warning("Configuring bluez permissions automatically...")
            if configure_bluez_conf_file():
                # We need to restart the service in this instance
                logger.warning(
                    "BlueZ permissions were modified, you must restart the bluetooth service for changes "
                    "to take affect. (Trying to restart now)"
                )
                subprocess.call(["service", "bluetooth", "restart"], shell=True)

    def __setattr__(self, key, value):
        self.__dict__[key] = value

        # If something is setting the error state, call recovery functions
        if key == "error_state" and value and self.auto_recover:
            logger.warning(
                "Error state declared by a module with auto recover set, attempting to recover"
            )
            try:
                for advertisement in self.advertisements.values():
                    asyncio.create_task(self.refresh_advertisement(advertisement))
                self.error_state = False
                logger.info("Recovered from error state!")
            except Exception:
                logger.exception("Recovery failed...")

    async def cleanup(self) -> None:
        """
        This function should be called when the program exits and will handle all cleanup similar to a
        destructor.

        TODO(Bernie): Debug handling here

        :return: None
        """

        if self.bus is not None:

            async def _cleanup():
                for rule in self.match_rules:
                    await remove_match(self.bus, match_rule=rule)

            asyncio.run(_cleanup())
            self.bus.disconnect()
        logger.info("BluezApplication cleanup completed. ")

    async def _get_proxy_object(self, bus_name: str, path: str) -> ProxyObject:
        """Gets a proxy object, the dbus equivalent is bus.get_object(bus_name, path)"""
        # Check if we already have this proxy stored
        if bus_name not in self.proxy_objects:
            self.proxy_objects[bus_name] = {}

        if path not in self.proxy_objects[bus_name]:
            self.proxy_objects[bus_name][path] = self.bus.get_proxy_object(
                bus_name, path, await self.bus.introspect(bus_name, path)
            )

        # Return the proxy object
        return self.proxy_objects[bus_name][path]

    async def get_proxy_interface(
        self, bus_name: DBusInterface, path: str, interface_name: DBusInterface
    ) -> ProxyInterface:
        """Gets a proxy interface using the _get_proxy_object method"""
        proxy_object = await self._get_proxy_object(bus_name.value, path)
        return proxy_object.get_interface(interface_name.value)

    async def setup(self) -> None:
        """
        Setup the application and get it on the dbus, this is a necessary function call to initialize the object.

        :return: None
        """
        # Get on the Dbus
        self._bus = await MessageBus(bus_type=BusType.SYSTEM).connect()
        self.bus.add_message_handler(self._dbus_message_parser)

        await self.bus.request_name(
            f"{BluezServiceInterfaces.BLUEZ_ORG.value}.{self.name}"
        )

        # Connect to bluez
        remote_object_manager = await self.get_proxy_interface(
            bus_name=BluezServiceInterfaces.BLUEZ_ORG,
            path="/",
            interface_name=DbusServiceInterfaces.OBJECT_MANAGER_INTERFACE,
        )

        # Find the path of an adapter with advertising capability
        if self.adapter_path is None:
            objects = await remote_object_manager.call_get_managed_objects()  # type: ignore
            for adapter, props in objects.items():
                if BluezServiceInterfaces.ADVERTISING_MANAGER_INTERFACE.value in props:
                    self.adapter_path = adapter
                    break

        # If no adapter is found, error out
        if self.adapter_path is None:
            raise RuntimeError("No adapter found, is a bluetooth adapter plugged in?")

        # Power the adapter
        logger.debug(self.adapter_path)
        self.adapter_properties = await self.get_proxy_interface(
            bus_name=BluezServiceInterfaces.BLUEZ_ORG,
            path=self.adapter_path,
            interface_name=DbusServiceInterfaces.PROPERTIES_INTERFACE,
        )
        await self.adapter_properties.call_set(  # type: ignore
            BluezServiceInterfaces.ADAPTER_INTERFACE.value,
            "Powered",
            Variant("b", True),
        )
        logger.debug(f"{self.adapter_path} is powered!")

        # Check if the adapter supports advertising while connected (ble v^5)
        self.always_advertising_supported = _check_support_for_always_advertising(
            self.adapter_path
        )
        if not self.always_advertising_supported:
            logger.warning(
                "HCI version is sub 5.0 which doesn't support advertising while connected always."
            )

        # Get the adapter interface, we need this for removing devices
        self.adapter_interface = await self.get_proxy_interface(
            bus_name=BluezServiceInterfaces.BLUEZ_ORG,
            path=self.adapter_path,
            interface_name=BluezServiceInterfaces.ADAPTER_INTERFACE,
        )

        # Find the advertisement manager and store it for later
        self.advertising_manager = await self.get_proxy_interface(
            bus_name=BluezServiceInterfaces.BLUEZ_ORG,
            path=self.adapter_path,
            interface_name=BluezServiceInterfaces.ADVERTISING_MANAGER_INTERFACE,
        )

        # Try to get the maximum number of advertisements we can use
        try:
            self.max_advertisements = (
                await self.advertising_manager.get_supported_instances()  # type: ignore
            )
            logger.info(f"Hardware supports {self.max_advertisements} advertisements.")
        except Exception:
            logger.exception(
                "Unable to get the number of supported instances from the adapter."
            )

        # Try to get the supported capabilities of advertising manager
        try:
            self.advertising_capabilities = (
                await self.advertising_manager.get_supported_capabilities()  # type: ignore
            )

            # Try to pull out variant values
            for key, value in self.advertising_capabilities.items():
                if isinstance(value, Variant):
                    self.advertising_capabilities[key] = value.value

            logger.info(
                f"Advertising manager supports: {self.advertising_capabilities}"
            )
        except Exception:
            logger.exception(
                "Unable to get supported capabilities of advertising manager."
            )

        # Try to get the params we can utilize
        try:
            self.supported_includes = (
                await self.advertising_manager.get_supported_includes()  # type: ignore
            )
            logger.info(
                f"Hardware supports {self.supported_includes} which can be controlled in software"
            )
        except Exception:
            logger.exception(
                "Unable to get the supported includes from the adapter interface."
            )

        logger.info(f"BluezApplication started with adapter: {self.adapter_path}")

        self.match_rules.add(
            "type=signal,interface=org.freedesktop.DBus.Properties,member=PropertiesChanged,"
            f"path_namespace={self.adapter_path}"
        )
        self.match_rules.add(
            "type=signal,interface=org.freedesktop.DBus.ObjectManager,member=InterfacesRemoved,"
            f"arg0path={self.adapter_path}/"
        )
        self.match_rules.add(
            "type=signal,interface=org.freedesktop.DBus.ObjectManager,member=InterfacesAdded,"
            f"arg0path={self.adapter_path}/"
        )

        for rule in self.match_rules:
            await add_match(self.bus, match_rule=rule)

        self.advertising_lock = asyncio.Lock()
        self.dbus_lock = asyncio.Lock()
        logger.debug("Setup complete!")

    async def add_advertisement(self, advertisement: "ABleAdvertisement") -> None:
        """Adds an le_advertisement object to the application"""

        # This needs to be under a lock on the case that multiple tasks are changing resources
        #   (ex. A task is switching ibeacons on the side)
        async with self.dbus_lock:
            # Fetch the bluez advert from the wrapper
            le_advertisement = advertisement.le_advertisement

            # Input validation
            if not isinstance(le_advertisement, BluezLEAdvertisement):
                raise TypeError(
                    f"Advertisement error is the wrong type for this platform (BlueZ): {type(le_advertisement)}"
                )

            if le_advertisement in self.advertisements.values():
                raise ValueError("Attempting to add le_advertisement already added!")

            if len(self.advertisements) == self.max_advertisements:
                raise RuntimeError("Unable to add any more advertisements!")

            # Get the path we will place it on the dbus, we need to find the first index we aren't actively using
            valid_indices = list(
                set(range(self.max_advertisements)) - set(self.advertisements.keys())
            )

            advertisement_index = valid_indices[0]
            advertisement_path = get_advertisement_object_path(
                self.name, advertisement_index
            )

            # Export the object onto the system bus
            self.bus.export(advertisement_path, le_advertisement)

            # Add it to bluez
            try:
                await self.advertising_manager.call_register_advertisement(  # type: ignore
                    advertisement_path, {}
                )
            except DBusError as e:
                raise RuntimeError(
                    "Attempted to add an invalid le_advertisement, maybe it is too long or the params are invalid, try"
                    "running the script again while watching $ sudo btmon"
                ) from e

            # Add this application to the le_advertisement
            le_advertisement.application = self

            # Add it to our tracker of advertisements
            self.advertisements[advertisement_index] = advertisement
            logger.debug(f"Advertisement added! ({advertisement_path})")

    async def remove_advertisement(self, advertisement: "ABleAdvertisement") -> None:
        """Removes an le_advertisement object from the application"""

        # See note in add advertisement for why this lock exists
        async with self.dbus_lock:
            # Input validation
            if not isinstance(advertisement.le_advertisement, BluezLEAdvertisement):
                raise TypeError(
                    f"Advertisement error is the wrong type for this platform: {type(advertisement.le_advertisement)}"
                )

            if advertisement not in self.advertisements.values():
                raise ValueError(
                    "Attempting remove le_advertisement not being advertised!"
                )

            # Get the path we will place it on the dbus
            advertisement_index = None
            for index, active_adv in self.advertisements.items():
                if active_adv == advertisement:
                    advertisement_index = index

            assert advertisement_index is not None, "This should be impossible..."
            advertisement_path = get_advertisement_object_path(
                self.name, advertisement_index
            )

            try:
                await self.advertising_manager.call_unregister_advertisement(  # type: ignore
                    advertisement_path
                )
                del self.advertisements[advertisement_index]

                # Export the object onto the system bus
                self.bus.unexport(advertisement_path, advertisement.le_advertisement)
            except Exception:
                logger.exception(
                    "Unable to remove le_advertisement! (this is okay during recovery)"
                )

            logger.debug(f"Advertisement removed ({advertisement_path})")

    def add_service(self, service: "ABleService") -> None:
        """
        Adds a bluez service to the application, this must be done before running start_application()

        :raises TypeError: if service is not of type BluezService
        :raises RuntimeError: if the service has already been added to the application

        :param service: the service to add to the bluez application
        :return: None
        """
        if not isinstance(service, BluezService):
            raise TypeError(
                f"The service that is trying to be added is the wrong type for this platform (BlueZ): {type(service)}"
            )

        # Double check it has not already been added
        if service in self.services:
            raise RuntimeError(f"Already added {service} to the application.")

        # Determine the path of the service and add it to the services list
        # pylint: disable=protected-access
        service._path = get_service_path(self.name, len(self.services) + 1)
        self.services.append(service)

        # Export this service as an object on the dbus
        self.bus.export(service.path, service)

    def add_characteristic(
        self,
        service: "ABleService",
        characteristic: "ABleCharacteristic",
        is_comms_char: bool = False,
    ) -> None:
        """
        Adds a characteristic to the application, this must done before running start_application()

        TODO(Bernie): What kind of validation should we do here?

        :param is_comms_char: if `True` the characteristic will become the comms characteristic, defaults to `False`
        :param service: the service which the characteristic will be added under
        :param characteristic: the characteristic you wish to add to the service
        :return: None
        """
        if not isinstance(service, BluezService):
            raise TypeError(
                "The service must be the of type BluezService to add to a Bluez application"
            )
        if not isinstance(characteristic, BluezCharacteristic):
            raise TypeError(
                "The characteristic must be of the type BluezCharacteristic to add to a Bluez application"
            )

        # Make sure there wasn't already a characteristic with that UUID
        if characteristic.uuid in service.characteristics.keys():
            raise ValueError(
                "Adding a characteristic with a UUID that matches another characteristic in the service!"
            )

        # Calculate the path of the characteristic
        characteristic._path = (  # pylint: disable=protected-access
            get_characteristic_path(service.path, len(service.characteristics) + 1)
        )

        # Add the characteristic to the service
        service.characteristics[characteristic.uuid] = characteristic

        # Link the characteristic to the service
        characteristic.service = service
        characteristic.application = self

        if is_comms_char:
            # Now handle marking updating by removing the old one and adding the new one
            if self.marked_characteristic is not None:
                logger.debug(
                    f"{self.marked_characteristic} is no longer marked for comms."
                )
                self.marked_characteristic.is_marked = False

            characteristic.is_marked = True
            self.marked_characteristic = characteristic
            logger.debug(f"{self.marked_characteristic} is now marked for comms")

        # Export it on the bus
        self.bus.export(characteristic.path, characteristic)

    def add_connect_callback(self, callback: Callable) -> None:
        """
        Adds a user-defined callback function to the application to be called after connecting to a central device.
        :param callback: the function to be called after a successful connection
        :return: None
        """
        self.user_connect_callback = callback

    def add_disconnect_callback(self, callback: Callable) -> None:
        """
        Adds a user-defined callback function to the application to be called after disconnecting from a central device.
        :param callback: the function to be called after a successful disconnect
        :return: None
        """
        self.user_disconnect_callback = callback

    async def start_interface(self):
        """
        Starts the GATT manager interface by registering it with the bluez service manager. Note that any
        services and their respective characteristics must be added before this is called.

        :return: None
        """
        # Check to see that at least one service was added, bluez will not allow willy nilly service-less advertising
        # https://github.com/luetzel/bluez/blob/aae6a2c4ce9963db110535647aa723b96561f6ac/src/gatt-database.c#L2452
        if len(self.services) == 0:
            raise RuntimeError(
                "You cannot start the interface on bluez without adding at least one service"
            )

        if self.gatt_manager_interface is None:
            self.gatt_manager_interface: ProxyInterface = (
                await self.get_proxy_interface(
                    BluezServiceInterfaces.BLUEZ_ORG,
                    self.adapter_path,
                    BluezServiceInterfaces.GATT_MANAGER_INTERFACE,
                )
            )

            # Determine if the bluez installed supported directed notifications
            self.supports_directed_notifications = (
                _check_support_for_directed_notifications(self.gatt_manager_interface)
            )

        # Try to unexport anything on the message bus
        self.bus.unexport(self.path)

        self.bus.export(self.path, self)
        await self.gatt_manager_interface.call_register_application(self.path, {})
        logger.debug("Registered application with the bluez service manager.")

    async def refresh_advertisement(self, advertisement: "ABleAdvertisement"):
        """Hack to try to update params by stopping and restarting the interface"""
        if not isinstance(self.advertising_lock, asyncio.Lock):
            raise RuntimeError(
                "Unable to refresh advertisement, has the setup method been ran?"
            )

        async with self.advertising_lock:
            await self.remove_advertisement(advertisement)
            await self.add_advertisement(advertisement)

    def _dbus_message_parser(self, message: Message) -> None:
        """
        Handler for all dbus messages that we have match rules for. We have two main routes and check them
        through either properties changed or interfaces added.

        TODO(Ori): Change the mac address parsing to use the parse_identifier_from_path function in utils.py

        :param message: The dbus message we caught.
        :return: None
        """
        try:
            if message.member == "PropertiesChanged":
                # Handle validation, we are expecting a list of length 3
                if not isinstance(message.body, list) or len(message.body) != 3:
                    return

                # The first element in the 3 long list needs to be org.bluez.Device1
                if message.body[0] != "org.bluez.Device1":
                    return

                # Make sure the message has a path
                if message.path is None:
                    return

                data: dict = message.body[1]
                object_path: str = message.path
                device_mac = object_path[
                    object_path.find("_") + 1 : object_path.rfind("_") + 3
                ]

                # Make sure that this is the connected properties changed signal
                if "Connected" not in data:
                    return

                if data["Connected"].value:
                    logger.info("Calling connected callback off of props changed...")
                    self.connect_callback(
                        DeviceCallbackData(devices_identifier=device_mac)
                    )
                else:
                    logger.info("Calling disconnect callback off of props changed...")
                    self.disconnect_callback(
                        DeviceCallbackData(devices_identifier=device_mac)
                    )

            elif message.member == "InterfacesAdded":
                # Handle validation, we expect a list of length 2
                if not isinstance(message.body, list) or len(message.body) != 2:
                    return

                # Make sure that a device path is the first element in the path
                if self.adapter_path not in message.body[0]:
                    return

                # Make sure this is the right interfaces added
                if "org.bluez.Device1" not in message.body[1]:
                    return

                # Make sure that this is a device with an address!
                if "Address" not in message.body[1]["org.bluez.Device1"]:
                    return

                # Make sure that it is connected and not a signal from a disconnected but cached device
                if (
                    "Connected" in message.body[1]["org.bluez.Device1"]
                    and not message.body[1]["org.bluez.Device1"]["Connected"].value
                ):
                    return

                # Parse out the path of the central
                central_mac = message.body[1]["org.bluez.Device1"]["Address"].value

                # Call the connect callback
                logger.info(
                    "Calling connected callback off of interfaces added signal..."
                )
                self.connect_callback(
                    DeviceCallbackData(devices_identifier=central_mac)
                )

        except Exception:
            logger.exception(
                "Unhandled exception when parsing dbus-next signal message..."
            )

    def connect_callback(self, callback_data: DeviceCallbackData) -> None:
        """
        Handler for the new connections being signalled from bluez. Verifies that the central was not already connected
        and is being stored in the connected_centrals dictionary (this would mean we lost state). Adds an entry to the
        new connection queue so that a peripheral server can accept the new connection. Adds a new ABleCentral to the
        connected centrals dictionary. Calls an optional user defined callback last.

        :param callback_data: data including the identifier required to create ABleCentral
        :return: None
        """
        from able.central import ABleCentral  # pylint: disable=import-outside-toplevel

        logger.info(f"New connection from {callback_data}")

        if callback_data.identifier in self.connected_centrals:
            raise RuntimeWarning(
                f"Received a double connection from {callback_data.identifier}, we missed a disconnect "
                "callback"
            )

        # Add to the NCQ
        new_central: "ABleCentral" = ABleCentral(
            application=self,
            identifier=callback_data.identifier,
            adapter_path=self.adapter_path,
        )
        self.new_connection_queue.add(new_central)
        self.connected_centrals[callback_data.identifier] = new_central

        # Add a receive queue for each characteristic
        for service in self.services:
            for char in service.characteristics.values():
                new_central.char_queues[hash(char)] = asyncio.Queue()
                logger.debug(
                    f"Queue was created for characteristic {char.uuid} on central {new_central}"
                )

        # Set the new connection event if it is not already set
        if not self.has_new_central.is_set():
            self.has_new_central.set()

        logger.debug(f"{self.new_connection_queue} | {self.connected_centrals}")

        if self.always_advertising_supported:
            for advert in self.advertisements.values():
                asyncio.create_task(self.refresh_advertisement(advert))

        if self.user_connect_callback:
            try:
                self.user_connect_callback(callback_data)
            except TypeError:
                logger.exception(
                    "User-defined connection callback is not able to be called with the argument given"
                )

    def disconnect_callback(self, callback_data: DeviceCallbackData) -> None:
        """
        Handler for disconnections being signalled by bluez. Updates the disconnected ABleCentral's state to be
        disconnected. Removes the central from the connected centrals dictionary (if it was there, it will not be
        present if the connection was established before the application was running) and the new connection queue
        if it is present. Calls an optional user defined callback last.

        :param callback_data: data regarding the central which disconnected
        :return: None
        """
        from able.central import ABleCentral  # pylint: disable=import-outside-toplevel

        logger.info(f"Disconnect from {callback_data}")

        # Make sure the central was already connected, if not it is okay
        if callback_data.identifier not in self.connected_centrals:
            logger.warning("A central disconnected which we did not know was connected")
            return

        # Get the central which disconnected the centrals state to disconnected
        disconnecting_central: ABleCentral = self.connected_centrals[
            callback_data.identifier
        ]
        disconnecting_central.disconnected_event.set()

        # Remove from the connected centrals dict and the NCQ
        self.new_connection_queue.discard(disconnecting_central)
        self.connected_centrals.pop(callback_data.identifier)

        # If we no longer have any connections, unset the event
        if len(self.new_connection_queue) == 0:
            logger.debug(
                "There are no more clients in the NCQ, clearing has_new_central event"
            )
            self.has_new_central.clear()

        logger.info(f"{self.new_connection_queue} | {self.connected_centrals}")

        if self.always_advertising_supported:
            for advert in self.advertisements.values():
                asyncio.create_task(self.refresh_advertisement(advert))

        if self.user_disconnect_callback:
            try:
                self.user_disconnect_callback(callback_data)
            except TypeError:
                logger.exception(
                    "User-defined disconnect callback is not able to be called with the argument given"
                )

    async def disconnect(self, central: "ABleCentral") -> None:
        """
        Call bluez to disconnect a connected central, this will not update the connected state of the central because
        the disconnect callback will handle that.

        # TODO(Bernie, Ori): Test the reliability of using remove_device vs disconnect

        :param central: The central object to force a disconnect.
        :return: None
        """
        logger.info(f"Disconnecting from central: {central.identifier}")

        try:
            # await self.adapter_interface.call_remove_device(central.dbus_path)
            device_interface = await self.get_proxy_interface(
                BluezServiceInterfaces.BLUEZ_ORG,
                central.dbus_path,
                BluezServiceInterfaces.DEVICE_INTERFACE,
            )

            await device_interface.call_disconnect()  # type: ignore
            logger.info(f"Disconnected from {central}...")
        except Exception:
            logger.exception(f"Unable to disconnect from {central.dbus_path}")

    @property
    def communication_characteristic(self) -> BluezCharacteristic:
        """Property which returns the marked comms characteristic or raises an exception"""
        if self.marked_characteristic is None:
            raise RuntimeError(
                "Attempted to fetch comms characteristic without defining one"
            )
        return self.marked_characteristic

    @property
    def bus(self) -> MessageBus:
        """Property which returns the bus of this application if it is defined"""
        if self._bus is None:
            raise RuntimeError(
                "Message bus of the application was set to None, application cannot operate"
                "without the bus"
            )
        return self._bus



"""Wrapper around CoreBluetooth characteristic"""
# Native dependencies
import logging
import time
from enum import Enum
from typing import TYPE_CHECKING, Optional, Union
from uuid import UUID

# Core Bluetooth dependencies
from CoreBluetooth import CBUUID, CBMutableCharacteristic  # type: ignore

# Able dependencies
from able.characteristic import ABleCharacteristic


class CBCharacteristicFlags(Enum):
    """CoreBluetooth characteristic properties"""

    BROADCAST = 0x0001
    READ = 0x0002
    WRITE_WITHOUT_RESPONSE = 0x0004
    WRITE = 0x0008
    NOTIFY = 0x0010
    INDICATE = 0x0020
    AUTHENTICATED_SIGNED_WRITES = 0x0040
    EXTENDED_PROPERTIES = 0x0080
    RELIABLE_WRITE = 0x0100
    WRITABLE_AUXILIARIES = 0x0200


class GATTAttributePermissions(Enum):
    """CoreBluetooth characteristic permissions"""

    READ = 0x1
    WRITE = 0x2
    READ_ENCRYPTION_REQUIRED = 0x4
    WRITE_ENCRYPTION_REQUIRED = 0x8


if TYPE_CHECKING:
    from able.central import ABleCentral
    from able.corebluetooth.application import CBApplication

logger = logging.getLogger(name=__name__)


class CBCharacteristic(ABleCharacteristic):
    """
    CoreBluetooth implementation of the BlessGATTCharacteristic

    :param characteristic_uuid: the uuid of the characteristic
    :type characteristic_uuid: Union[str, UUID]
    :param value: the initial value of the characteristic
    :param flags: the characteristic properties (read, write, notify, etc.)
    :param permissions: the characteristic permissions (read, write, read_encryption_required, etc.)
    :param notify_on_write: if True, all subscribed centrals will be notified when a write occurs
    """

    def __init__(
        self,
        characteristic_uuid: Union[str, UUID],
        value=None,
        flags=None,
        permissions=None,
        nickname: str = None,
        notify_on_write: bool = False,
    ):
        super().__init__(characteristic_uuid, value, flags, nickname, notify_on_write)

        # Additional Members
        self.application: Optional["CBApplication"] = None
        self.permissions = permissions

        # Keep track of whether or not this characteristic is the char for socket comms
        self.is_marked = False

        logger.debug(f"Setting up characteristic with flags: {self.flags}")

        # Convert flags to properties and permissions
        self.process_flags()

        # Initialize the Core Bluetooth object
        cb_uuid: CBUUID = CBUUID.alloc().initWithString_(self.uuid)
        cb_characteristic: CBMutableCharacteristic = (
            CBMutableCharacteristic.alloc().initWithType_properties_value_permissions_(
                cb_uuid, self.properties_value, self.value, self.permissions_value
            )
        )
        self.obj = cb_characteristic
        self.last_write = time.time()

    def __str__(self):
        return (
            f"CBCharacteristic [{self.uuid}]"
            if self.nickname is None
            else self.nickname
        )

    def __repr__(self):
        return str(self)

    def process_flags(self) -> None:
        """Convert characteristic flags to Core Bluetooth properties and permissions"""
        self.properties_value = sum(self.flags)

        # Permissions are inferred from properties UNLESS user sepcifies them
        if not self.permissions:
            self.permissions = []
            property_names = [x.name for x in self._flags]
            if "READ" in property_names:
                self.permissions.append(GATTAttributePermissions.READ.value)
            if "WRITE" in property_names:
                self.permissions.append(GATTAttributePermissions.WRITE.value)

        self.permissions_value = sum(self.permissions)

    async def broadcast(self, data: Union[bytes, str]):
        """Notify every subscribed central of a change to this characteristic

        :param data: the data to notify the centrals on
        """
        if self.application.connected_centrals:  # type: ignore
            await self.notify_central_of_change(None, data, notify_all=True)
        else:
            self.value = data  # type: ignore
            self.obj.setValue_(data)

    async def notify_central_of_change(  # pylint: disable=arguments-differ
        self,
        central: Optional["ABleCentral"],
        data: Union[bytes, str],
        notify_all=False,
    ) -> None:
        """
        Makes a Core Bluetooth call to notify only one central of a change to this characteristic

        :param notify_all: Notify every connected device?
        :param central: the central object that is notified
        :param data: the data to 'send' to the central by setting the value of this characteristic to and
            notifying only that central
        :return: None
        """
        # now = time.time()
        # if now - self.last_write < .001:
        #     await asyncio.sleep(.001)
        # self.last_write = now

        if not self.notifying:
            raise RuntimeWarning(
                "No central has indicated they want notifications on this characteristic, a central "
                "must subscribe before notifying/indicating"
            )

        if (
            self.application is None
            or self.application.peripheral_manager_delegate is None
        ):
            raise RuntimeError(
                "Unable to notify central of change without a bound application"
            )

        peripheral_manager = (
            self.application.peripheral_manager_delegate.peripheral_manager
        )

        try:
            # TODO(Chad): Make sure this cannot get stuck in an infinite loop
            while not peripheral_manager.updateValue_forCharacteristic_onSubscribedCentrals_(
                data, self.obj, None if notify_all else [central.obj]  # type: ignore
            ):
                logger.debug(
                    "Could not send data... peripheral delegate manager cache full... trying again in a bit"
                )
                self.application.peripheral_manager_delegate.peripheral_ready.clear()
                self.application.peripheral_manager_delegate.peripheral_ready.wait(
                    timeout=0.0001
                )
        except Exception:
            logger.exception(
                f"Exception caught when trying to notify {central}, are you sure the central is setup"
                "to notify on the characteristic?"
            )

        logger.debug("successfully sent data to central(s)")
        self.value = data  # type: ignore
        self.obj.setValue_(data)

    def set_value(self, value: bytearray, central_identifier: str) -> None:
        """Handles the request over CoreBluetooth to write a value to a characteristic

        :param value: value received from CoreBluetooth write callback
        :param central_identifier: uuid of the central device that wrote
        """

        if self.application is None:
            raise RuntimeError(
                "No application bound for characteristic, unable to perform actions"
            )

        if central_identifier not in self.application.connected_centrals:
            logger.warning(
                f"Received a write from {central_identifier} who we aren't tracking, ignoring"
            )
            return

        # Add the data to the recv queue
        logger.debug(f"Putting data {value} on queue")
        writer: "ABleCentral" = self.application.connected_centrals[central_identifier]

        writer.char_queues[hash(self)].put_nowait(bytes(value))

        logger.debug("Updating value of characteristic")
        self.value = bytes(value)
        self.obj.setValue_(value)

        if self.notify_on_write:
            self.broadcast(value)



"""CoreBluetooth application that calls into the PeripheralManagerDelegate"""

#    Copyright (c)  2021  Allthenticate

# Native Libraries
import asyncio
import logging
import struct
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Set

# Core Bluetooth related
from CoreBluetooth import (  # type: ignore
    CBUUID,
    CBAdvertisementDataLocalNameKey,
    CBAdvertisementDataServiceUUIDsKey,
    CBCentral,
    CBManagerStatePoweredOff,
    CBManagerStatePoweredOn,
    CBMutableCharacteristic,
)
from Foundation import NSData, NSString  # type: ignore

# Able related
from able.corebluetooth.advertisement import CBLEAdvertisement
from able.corebluetooth.characteristic import CBCharacteristic
from able.corebluetooth.peripheral_manager_delegate import (
    PeripheralManagerDelegate,  # type: ignore
)
from able.corebluetooth.service import ABleCBService
from able.plugins.ibeacon import ABleiBeacon
from able.utils import DeviceCallbackData

if TYPE_CHECKING:
    from able.advertisement import ABleAdvertisement
    from able.central import ABleCentral
    from able.characteristic import ABleCharacteristic
    from able.service import ABleService

logger = logging.getLogger(name=__name__)


class CBApplication:
    """
    The high level application that abstracts the pyobjc corebluetooth calls

    :param name: the name that this application has
    """

    def __init__(self, name: str, *args, **kwargs):  # pylint: disable=unused-argument
        """Initializer for the CoreBluetooth application"""
        self.name: str = name

        self.peripheral_manager_delegate: PeripheralManagerDelegate = (
            PeripheralManagerDelegate.alloc().init()
        )

        # Allow the peripheral manager delegate to reference the app
        self.peripheral_manager_delegate.assign_app(self)

        # Services related
        self.services: List[ABleCBService] = []

        # Adverts related
        self.advertisement_data: Dict[
            CBAdvertisementDataLocalNameKey, CBAdvertisementDataServiceUUIDsKey
        ] = {}

        # Communication related
        self.marked_characteristic: Optional["ABleCharacteristic"] = None

        # Connection/subscription related
        self.has_new_central = asyncio.Event()
        self.new_connection_queue: Set["ABleCentral"] = set()
        self.connected_centrals: Dict[str, "ABleCentral"] = {}
        self.user_connect_callback: Optional[Callable] = None
        self.user_disconnect_callback: Optional[Callable] = None

        # Other
        self.start_counter = 0

        self.peripheral_ready = asyncio.Event()

    async def setup(self) -> None:
        """Setup the peripheral manager delegate and give it a reference to this app"""
        await self.peripheral_manager_delegate.setup()

    def state_change_callback(self, peripheral_state: int) -> None:
        """
        This is called anytime the state changes and will then trigger a user-defined callback

        :param peripheral_state: The current CoreBluetooth state of the Bluetooth controller
        :return: None
        """
        if peripheral_state == CBManagerStatePoweredOff:
            # Bluetooth off
            pass
        elif peripheral_state == CBManagerStatePoweredOn:
            # Bluetooth on
            logger.debug(
                "Trying to restart our advertisements (should force even when coming out of sleep)"
            )
            self.start_counter = 0
            asyncio.create_task(self.restart_advertising())

    def connect_callback(
        self, central: CBCentral, callback_data: DeviceCallbackData
    ) -> None:
        """
        Handler for the new connections/subscriptions being signalled from Core Bluetooth.
        Core Bluetooth does not offer an API for connection callbacks, just characteristic
        subscription callbacks. For the sake of cross-platform BLE abstraction, we treat the
        subscriptions the same as connections. Adds an entry to the new connection queue so
        that a peripheral server can accept the new connection. Adds a new ABleCentral to
        the connected centrals dictionary. Calls an optional user defined callback last.

        :param callback_data: data including the identifier required to create ABleCentral
        :return: None
        """
        from able.central import ABleCentral  # pylint: disable=import-outside-toplevel

        if callback_data.identifier in self.connected_centrals:
            logger.warning(
                "Client already connected, likely they already subscribed, connect cb is a NOOP"
            )
            return

        logger.info(f"New connection from {callback_data}")

        # Create the new central
        new_central: "ABleCentral" = ABleCentral(
            application=self, identifier=callback_data.identifier, central_obj=central
        )
        self.new_connection_queue.add(new_central)
        self.connected_centrals[callback_data.identifier] = new_central
        new_central.char_subscriptions.append(callback_data.char_uuid)

        # Add a receive queue for each characteristic
        for service in self.services:
            for char in service.characteristics.values():
                new_central.char_queues[hash(char)] = asyncio.Queue()
                logger.debug(
                    f"Queue was created for characteristic {char.uuid} on central {new_central}"
                )

        # Set the new connection event if it is not already set
        if not self.has_new_central.is_set():
            self.has_new_central.set()

        logger.debug(f"{self.new_connection_queue} | {self.connected_centrals}")

        if self.user_connect_callback:
            try:
                self.user_connect_callback(callback_data)
            except TypeError:
                logger.exception(
                    "User-defined connection callback is not able to be called with the argument given"
                )

    def disconnect_callback(self, callback_data: DeviceCallbackData) -> None:
        """
        Handler for disconnections being signalled by Core Bluetooth. Updates the disconnected ABleCentral's state to be
        disconnected. Removes the central from the connected centrals dictionary (if it was there, it will not be
        present if the connection was established before the application was running) and the new connection queue
        if it is present. Calls an optional user defined callback last.

        :param callback_data: data regarding the central which disconnected
        :return: None
        """

        from able.central import ABleCentral  # pylint: disable=import-outside-toplevel

        logger.info(f"Disconnect from {callback_data}")

        # Get the central which disconnected the centrals state to disconnected
        disconnecting_central: ABleCentral = self.connected_centrals[
            callback_data.identifier
        ]
        disconnecting_central.disconnected_event.set()

        # Remove from the connected centrals dict and the NCQ
        self.new_connection_queue.discard(disconnecting_central)
        self.connected_centrals.pop(callback_data.identifier)

        # If we no longer have any connections, unset the event
        if len(self.new_connection_queue) == 0:
            self.has_new_central.clear()

        logger.info(f"{self.new_connection_queue} | {self.connected_centrals}")

        if self.user_disconnect_callback:
            try:
                self.user_disconnect_callback(callback_data)
            except TypeError:
                logger.exception(
                    "User-defined disconnect callback is not able to be called with the argument given"
                )

    async def add_advertisement(
        self, advertisement: "ABleAdvertisement", refresh: bool = False
    ) -> None:
        """Adds an le_advertisement object to the application"""

        # Fetch the core bluetooth advert from the wrapper
        le_advertisement = advertisement.le_advertisement

        # Input validation
        if not isinstance(le_advertisement, CBLEAdvertisement):
            raise TypeError(
                f"Advertisement error is the wrong type for this platform (CoreBluetooth): {type(le_advertisement)}"
            )

        if self.advertisement_data and not refresh:
            raise ValueError(
                "There is already an advertisement in the application (CoreBluetooth supports only 1)"
            )

        # Add the advertisement to the application
        uuids: List[CBUUID] = [
            CBUUID.alloc().initWithString_(service_uuid)
            for service_uuid in le_advertisement.service_uuids
        ]
        self.advertisement_data = {
            CBAdvertisementDataLocalNameKey: le_advertisement.local_name,
            CBAdvertisementDataServiceUUIDsKey: uuids,
        }

        le_advertisement.application = self

    async def advertise_ibeacon(self, advertisement: ABleiBeacon) -> None:
        """
        Adds an iBeacon and then advertises it

        TODO: How many of these can a peripheral support?
        TODO (Chad): Figure out a more general way to do this in the future

        Ref: https://github.com/watr/mbeacon
        Ref: https://titanwolf.org/Network/Articles/Article?AID=b0003beb-d37f-4d08-a62f-16e1e93cc19b
        Ref: https://blendedcocoa.com/blog/2013/11/02/mavericks-as-an-ibeacon/
        """
        # I don't think we need this, 0 sounds fine
        measured_power = 0

        # We must convert everything into a RAW data buffer (this functionality is not supported/documented by Apple)
        raw_data = (
            advertisement.uuid.bytes
            + b"".join(advertisement.major)
            + b"".join(advertisement.minor)
            + struct.pack("B", measured_power)
        )

        # Convert our data to NSString and NSData so that the API is happy
        beacon_nsstring = NSString.alloc().initWithString_("kCBAdvDataAppleBeaconKey")
        data = NSData.alloc().initWithBytes_length_(raw_data, len(raw_data))
        beacon_name = NSString.alloc().initWithString_("aBeacon")

        # Construct our dictionary and advertise it!
        adv_data = {
            CBAdvertisementDataLocalNameKey: beacon_name,
            beacon_nsstring: data,
        }
        await self.peripheral_manager_delegate.startAdvertising_(adv_data)
        logger.info("Ibeacon added!")

    async def refresh_advertisement(self, advertisement: "ABleAdvertisement") -> None:
        """Refreshes the le_advertisement object and calls startAdvertising again with the new
        advertisement data"""
        self.peripheral_manager_delegate.stopAdvertising()
        await self.add_advertisement(advertisement, refresh=True)
        await self.restart_advertising()

    async def restart_advertising(self) -> bool:
        """Simply re-advertise our existing advertisement data.  This is useful when coming out sleep mode."""
        if self.advertisement_data is None:
            return False

        self.start_counter += 1
        try:
            await self.peripheral_manager_delegate.startAdvertising_(
                self.advertisement_data
            )
            self.start_counter = 0
            return True
        except asyncio.TimeoutError as e:
            logger.debug("Failed to start advertising, retrying...")
            if self.start_counter < 10:
                await self.start_interface()
            else:
                raise RecursionError(
                    "Failed to start advertising for 10 consecutive times, ending execution..."
                ) from e
            return False

    async def start_interface(self) -> None:
        """
        Starts up the CoreBluetooth interface

        :return: None
        """
        logger.debug(f"Advertisement Data: {self.advertisement_data}")

        for service in self.services:
            logger.debug(f"Adding service: {service.uuid}")
            await self.peripheral_manager_delegate.addService(service.obj)

        self.start_counter += 1
        try:
            await self.peripheral_manager_delegate.startAdvertising_(
                self.advertisement_data
            )
            self.start_counter = 0
        except asyncio.TimeoutError as e:
            logger.debug("Failed to start advertising, retrying...")
            if self.start_counter < 10:
                await self.start_interface()
            else:
                raise RecursionError(
                    "Failed to start advertising for 10 consecutive times, ending execution..."
                ) from e

    def add_service(self, service: "ABleCBService") -> None:
        """
        Adds a CoreBluetooth service to the application, this must be done before running start_application()

        :raises TypeError: if service is not of type ABleCBService
        :raises RuntimeError: if the service has already been added to the application

        :param service: the service to add to the CB application
        :return: None
        """
        if not isinstance(service, ABleCBService):
            raise TypeError(
                f"The service that is trying to be added is the wrong type for this platform (CoreBluetooth):"
                f" {type(service)}"
            )

        if service in self.services:
            raise RuntimeError(f"Already added {service} to the application.")

        self.services.append(service)

    def add_characteristic(
        self,
        service: "ABleService",
        characteristic: CBCharacteristic,
        is_comms_char: bool = False,
    ) -> None:
        """Adds a characteristic to the application

        :param is_comms_char: if `True` the characteristic will become the comms characteristic, defaults to `False`
        :param service: the service which the characteristic will be added under
        :param characteristic: the characteristic you wish to add to the service
        :return: None
        """

        if not isinstance(service, ABleCBService):
            raise TypeError(
                "The service must be the of type ABleCBService to add to a Bluez application"
            )
        if not isinstance(characteristic, CBCharacteristic):
            raise TypeError(
                "The characteristic must be of the type BluezCharacteristic to add to a Bluez application"
            )

        # Make sure there wasn't already a characteristic with that UUID
        if characteristic.uuid in service.characteristics.keys():
            raise ValueError(
                "Adding a characteristic with a UUID that matches another characteristic in the service!"
            )

        # Add the characteristic to the service
        service.characteristics[characteristic.uuid] = characteristic

        # Link the characteristic to the service
        characteristic.service = service
        characteristic.application = self

        if is_comms_char:
            # Now handle marking updating by removing the old one and adding the new one
            if self.marked_characteristic is not None:
                logger.debug(
                    f"{self.marked_characteristic} is no longer marked for comms."
                )
                self.marked_characteristic.is_marked = False  # type: ignore

            characteristic.is_marked = True  # type: ignore
            self.marked_characteristic = characteristic
            logger.debug(f"{self.marked_characteristic} is now marked for comms")

        # Add the characteristic to the Core Bluetooth service
        chars: List[CBMutableCharacteristic] = [
            char.obj for char in service.characteristics.values()  # type: ignore
        ]
        service.obj.setCharacteristics_(chars)

    def stop(self) -> None:
        """Stop the peripheral server"""
        self.peripheral_manager_delegate.peripheral_manager.removeAllServices()
        self.peripheral_manager_delegate.stopAdvertising()

    def add_connect_callback(self, callback: Callable) -> None:
        """
        Adds a user-defined callback function to the application to be called after connecting to a central device.
        :param callback: the function to be called after a successful connection
        """
        self.user_connect_callback = callback

    def add_disconnect_callback(self, callback: Callable) -> None:
        """
        Adds a user-defined callback function to the application to be called after disconnecting from a central device.
        :param callback: the function to be called after a successful disconnect
        """
        self.user_disconnect_callback = callback

    async def disconnect(
        self, central: "ABleCentral"
    ):  # pylint: disable=unused-argument, no-self-use
        """Disconnect from central"""
        logger.warning("Serverside disconnect is not currently supported on macOS")

    @property
    def communication_characteristic(self) -> "ABleCharacteristic":
        """Property which returns the marked comms characteristic or raises an exception"""
        if self.marked_characteristic is None:
            raise RuntimeError(
                "Attempted to fetch comms characteristic without defining one"
            )
        return self.marked_characteristic



#    Copyright (c)  2021  Allthenticate
"""
Abstraction of the ABleService that all backend services will inherit
"""

# Native modules
import uuid
from typing import TYPE_CHECKING, Dict, Union

if TYPE_CHECKING:
    from able.characteristic import ABleCharacteristic


class ABleService:
    """
    This is an abstract class for the Services that each platform backend defines. This class should never be used
    directly; instead, when the `ABleService` is imported from `able`, the appropriate subclass will be fetched
    based on what platform is being used.

    :param service_uuid: the uuid of the service.
    :ivar uuid: the uuid of the service
    :ivar characteristics:  a dictionary of uuid string - characteristic pairs, this is used to fetch a characteristic
        from a service if a reference to the characteristic is lost in your program.
    """

    def __init__(self, service_uuid: Union[str, uuid.UUID], **kwargs):
        """Initializes the able service"""

        # Define members used across all ABleServices
        self.uuid = str(service_uuid)
        self.characteristics: Dict[str, "ABleCharacteristic"] = {}

        # Work as a mixin
        super().__init__(**kwargs)  # type: ignore



"""
PyObjC object of CBPeripheralManagerDelegate

API reference:
https://developer.apple.com/documentation/corebluetooth/cbperipheralmanagerdelegate
"""
# pylint: skip-file

# Native libraries
import asyncio
import logging
import threading
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Set

import objc  # type: ignore
from CoreBluetooth import (  # type: ignore
    CBManagerStatePoweredOff,
    CBManagerStatePoweredOn,
    CBManagerStateResetting,
    CBManagerStateUnauthorized,
    CBManagerStateUnknown,
    CBManagerStateUnsupported,
)
from Foundation import (  # type: ignore
    CBATTRequest,
    CBCentral,
    CBCharacteristic,
    CBMutableService,
    CBPeripheralManager,
    CBService,
    NSError,
    NSObject,
)
from libdispatch import DISPATCH_QUEUE_SERIAL, dispatch_queue_create  # type: ignore

from able.corebluetooth.error import CBATTError

# Able Dependencies
from able.utils import DeviceCallbackData

if TYPE_CHECKING:
    from able import ABleCharacteristic
    from able.advertisement import ABleAdvertisement
    from able.central import ABleCentral
    from able.corebluetooth.application import CBApplication
    from able.corebluetooth.characteristic import CBCharacteristic

# Setup logging
logger = logging.getLogger(name=__name__)


class PeripheralManagerDelegate(NSObject):  # pylint: disable=invalid-name
    """Handles all the direct calls to the CBPeripheralManager"""

    CBPeripheralManagerDelegate = objc.protocolNamed("CBPeripheralManagerDelegate")
    ___pyobjc_protocols__ = [CBPeripheralManagerDelegate]

    def __init__(self):
        self._app = None

    async def setup(self) -> None:
        """
        Setup the CoreBluetooth peripheral manager

        :return: None
        """

        self = objc.super(  # pylint: disable=self-cls-assignment
            PeripheralManagerDelegate, self
        ).init()
        self.event_loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
        self.peripheral_manager: CBPeripheralManager = (
            CBPeripheralManager.alloc().initWithDelegate_queue_(
                self,
                dispatch_queue_create(b"bleak.corebluetooth", DISPATCH_QUEUE_SERIAL),
            )
        )

        self._callbacks: Dict[str, Callable] = {}

        # Events
        self._powered_on_event: threading.Event = threading.Event()
        self._advertisement_started_event: asyncio.Event = asyncio.Event()
        self._services_added_events: Dict[str, asyncio.Event] = {}
        self.peripheral_ready: threading.Event = threading.Event()

        # Wait for bluetooth module to be powered on
        self._powered_on_event.wait()

        if not self.compliant():
            logger.warning("PeripheralManagerDelegate is not compliant")

    @objc.python_method
    def assign_app(self, app: "CBApplication") -> None:
        """Method to assign the CBApplication to instances of this class"""
        self._app = app

    def compliant(self) -> bool:
        """
        Determines whether the class adheres to the CBPeripheralManagerDelegate
        protocol
        Returns
        -------
        bool
            Whether the class is compliant with the CBPeripheralManagerDelegate
            Protocol
        """
        return PeripheralManagerDelegate.pyobjc_classMethods.conformsToProtocol_(
            self.CBPeripheralManagerDelegate
        )

    def peripheralManagerDidUpdateState_(
        self, peripheral_manager: CBPeripheralManager
    ):  # pylint: disable=invalid-name
        """CB callback"""
        if peripheral_manager.state() == CBManagerStateUnknown:
            logger.debug("Cannot detect bluetooth device")
        elif peripheral_manager.state() == CBManagerStateResetting:
            logger.debug("Bluetooth is resetting")
        elif peripheral_manager.state() == CBManagerStateUnsupported:
            logger.debug("Bluetooth is unsupported")
        elif peripheral_manager.state() == CBManagerStateUnauthorized:
            logger.debug("Bluetooth is unauthorized")
        elif peripheral_manager.state() == CBManagerStatePoweredOff:
            logger.debug("Bluetooth powered off")
        elif peripheral_manager.state() == CBManagerStatePoweredOn:
            logger.debug("Bluetooth powered on")

        if peripheral_manager.state() == CBManagerStatePoweredOn:
            self._powered_on_event.set()
        else:
            self._powered_on_event.clear()
            self._advertisement_started_event.clear()

        self.event_loop.call_soon_threadsafe(
            self._app.state_change_callback, peripheral_manager.state()
        )

    def peripheralManager_willRestoreState_(  # pylint: disable=invalid-name,no-self-use
        self,
        peripheral: CBPeripheralManager,  # pylint: disable=unused-argument
        d: dict,
    ):
        """CB callback"""
        logger.debug("PeripheralManager restoring state: {}".format(d))

    ###############################
    # ADVERTISEMENT RELATED
    ###############################

    async def startAdvertising_(  # pylint: disable=invalid-name
        self,
        advertisement_data: Dict,
        timeout: float = 2.0,
    ):
        """
        Begin Advertising on the server
        Parameters
        ----------
        advertisement_data : Dict[str, Any]
            Dictionary of additional data to advertise. See Apple Documentation
            for more info
        timeout : float
            How long to wait before throwing an error if advertising doesn't
            start
        """

        self.peripheral_manager.startAdvertising_(advertisement_data)

        await asyncio.wait_for(self._advertisement_started_event.wait(), timeout)

        logger.debug(
            "Advertising started with the following data: {}".format(advertisement_data)
        )

    def is_advertising(self) -> bool:
        """
        Determine whether the server is advertising
        Returns
        -------
        bool
            True if advertising
        """
        return self.peripheral_manager.isAdvertising()

    def stopAdvertising(self):  # pylint: disable=invalid-name
        """
        Stop Advertising
        """
        self.peripheral_manager.stopAdvertising()

    @objc.python_method
    def peripheralManagerDidStartAdvertising_error(  # pylint: disable=invalid-name
        self, peripheral_manager: CBPeripheralManager, error: NSError
    ):
        """CB callback"""
        if error:
            raise Exception("Failed to start advertising: {}".format(error))

        logger.debug("Peripheral manager did start advertising")
        self._advertisement_started_event.set()

    def peripheralManagerDidStartAdvertising_error_(  # pylint: disable=invalid-name
        self, peripheral_manager: CBPeripheralManager, error: NSError
    ):
        """CB callback"""
        logger.debug("Received DidStartAdvertising Message")
        self.event_loop.call_soon_threadsafe(
            self.peripheralManagerDidStartAdvertising_error, peripheral_manager, error
        )

    ###############################
    # SERVICE RELATED
    ###############################

    @objc.python_method
    async def addService(  # pylint: disable=invalid-name
        self, service: CBMutableService
    ):  # pylint: disable=invalid-name
        """
        Add a service to the peripheral

        Parameters
        ----------
        service : CBMutableService
            The service to be added to the server
        """
        service_uuid: str = service.UUID().UUIDString()
        self._services_added_events[service_uuid] = asyncio.Event()

        self.peripheral_manager.addService_(service)

        await self._services_added_events[service_uuid].wait()

    @objc.python_method
    def peripheralManager_didAddService_error(  # pylint: disable=invalid-name
        self,
        peripheral_manager: CBPeripheralManager,
        service: CBService,
        error: NSError,
    ):
        """CB callback"""
        uuid: str = service.UUID().UUIDString()
        if error:
            raise Exception("Failed to add service {}: {}".format(uuid, error))

        logger.debug("Peripheral manager did add service: {}".format(uuid))
        logger.debug(
            "service added had characteristics: {}".format(service.characteristics())
        )
        self._services_added_events[uuid].set()

    def peripheralManager_didAddService_error_(  # pylint: disable=invalid-name
        self,
        peripheral_manager: CBPeripheralManager,
        service: CBService,
        error: NSError,
    ):
        """CB callback"""
        self.event_loop.call_soon_threadsafe(
            self.peripheralManager_didAddService_error,
            peripheral_manager,
            service,
            error,
        )

    ###############################
    # CHARACTERISTIC RELATED
    ###############################

    def peripheralManager_central_didSubscribeToCharacteristic_(  # pylint: disable=invalid-name
        self,
        peripheral_manager: CBPeripheralManager,
        central: CBCentral,
        characteristic: CBCharacteristic,
    ):
        """CB callback"""
        central_uuid: str = central.identifier().UUIDString()
        char_uuid: str = characteristic.UUID().UUIDString()
        logger.debug(
            "Central Device: {} is subscribing to characteristic {}".format(
                central_uuid, char_uuid
            )
        )

        # Mark the characteristic as notifying
        char = self._get_char_from_uuid(char_uuid)
        if char is not None:
            char.client_subscribed.set()
        else:
            logger.debug(
                "Could not find characteristic that central subscribed to. Something bad must have happened!"
            )
            return

        # Check if this central is a new connection
        if central_uuid in self._app.connected_centrals:
            logger.debug(f"Central Device {central_uuid} is already connected")
            # Add the char UUID to central's char subscriptions
            central = self._app.connected_centrals[central_uuid]
            central.char_subscriptions.append(char_uuid)
        else:
            # Call connect callback in application
            device_data = DeviceCallbackData(central_uuid, char_uuid=char_uuid)
            self.event_loop.call_soon_threadsafe(
                self._app.connect_callback, central, device_data
            )

    def peripheralManager_central_didUnsubscribeFromCharacteristic_(  # pylint: disable=invalid-name
        self,
        peripheral_manager: CBPeripheralManager,
        central: CBCentral,
        characteristic: CBCharacteristic,
    ):
        """CB callback"""
        central_uuid: str = central.identifier().UUIDString()
        char_uuid: str = characteristic.UUID().UUIDString()
        logger.debug(
            "Central device {} is unsubscribing from characteristic {}".format(
                central_uuid, char_uuid
            )
        )

        # Mark the characteristic as NOT notifying
        char: "ABleCharacteristic" = self._get_char_from_uuid(char_uuid)
        if char is not None:
            char.client_subscribed.clear()
        else:
            logger.debug(
                "Could not find characteristic that central unsubscribed from. Something bad must have happened!"
            )
            return

        # Get central object
        central: "ABleCentral" = self._app.connected_centrals.get(central_uuid)

        # Make sure this is actually a central that we manage
        if central is None:
            logger.warning(
                "Could not find central (%s) in an unsubscribe!" % central_uuid
            )
            return

        # Remove char from subscriptions list
        try:
            central.char_subscriptions.remove(char_uuid)
        except Exception:
            logger.debug(f"oops! couldn't find the characteristic {char_uuid}")

        # Check if central is not subscribed to any chars, if so call disconnect callback
        if not central.char_subscriptions:
            device_data = DeviceCallbackData(central_uuid, char_uuid=char_uuid)
            self.event_loop.call_soon_threadsafe(
                self._app.disconnect_callback, device_data
            )

    def peripheralManagerIsReadyToUpdateSubscribers_(  # pylint: disable=invalid-name
        self, peripheral_manager: CBPeripheralManager
    ):
        """CB callback"""
        logger.debug("Peripheral is ready to update subscribers")
        self.peripheral_ready.set()

    def peripheralManager_didReceiveWriteRequests_(  # pylint: disable=invalid-name
        self, peripheral_manager: CBPeripheralManager, requests: List[CBATTRequest]
    ):
        """CB callback"""
        logger.debug("Receving write requests...")
        for request in requests:
            central_uuid: CBCentral = request.central().identifier().UUIDString()
            char_uuid: CBCharacteristic = request.characteristic().UUID().UUIDString()
            value: bytearray = request.value()

            logger.debug(f"Value received: {value}")

            able_char: "CBCharacteristic" = self._get_char_from_uuid(char_uuid)
            self.event_loop.call_soon_threadsafe(
                able_char.set_value, value, central_uuid
            )

        peripheral_manager.respondToRequest_withResult_(
            requests[0], CBATTError.Success.value
        )

    def peripheralManager_didReceiveReadRequest_(  # pylint: disable=invalid-name
        self, peripheral_manager: CBPeripheralManager, request: CBATTRequest
    ):
        """CB callback"""
        logger.debug(
            "Received read request from {} for characteristic {}".format(
                request.central().identifier().UUIDString(),
                request.characteristic().UUID().UUIDString(),
            )
        )
        request.setValue_(request.characteristic().value())
        logger.debug(f"Reading {request.characteristic().value()}")
        peripheral_manager.respondToRequest_withResult_(
            request, CBATTError.Success.value
        )

    @objc.python_method
    def _get_char_from_uuid(self, char_uuid: str) -> Optional["CBCharacteristic"]:
        for service in self._app.services:
            try:
                return service.characteristics[char_uuid.lower()]
            except Exception:
                continue
        return None



#    Copyright (c)  2021  Allthenticate
"""
Implementation of a BluezDbus BLE advertisement using dbus_next as the interface with Bluez over the dbus system bus.
"""

# Native libraries
import logging
import uuid
from typing import Union

# Dependencies
from dbus_next import Variant  # type: ignore
from dbus_next.constants import PropertyAccess  # type: ignore
from dbus_next.service import ServiceInterface, dbus_property, method  # type: ignore

# Able Dependencies
from able.bluezdbus import AdvertisementType, BluezServiceInterfaces

# Setup logging
logger = logging.getLogger(name=__name__)


class BluezLEAdvertisement(ServiceInterface):
    """
    Class implementation of the BLE Advertisement Service through Dbus-Next, this is specific to the Linux platform and
    serves as the low level handler for individual advertisements.

    Note that right now we have not implemented a check for the advertisement's length so if you see errors in trying to
    add an advertisement, try removing some fields and try again, you may be over the max advertising length for your
    hardware.

    Note that the advertisement will also try to be restored if it is detected that the advertisement was released,
    triggering the application that holds it to refresh the advertisement in an attempt to recover it.

    :param local_name: The name you wish to include within advertisement packets as the local name, defaults to `BLE
        Peripheral`
    :type local_name: str
    :param advertisement_type: What type of advertisement this is, see `AdvertisementType` for the available types.
    :type advertisement_type: `AdvertisementType`, optional
    """

    def __init__(
        self,
        advert_wrapper,
        local_name="BLE Peripheral",
        advertisement_type=AdvertisementType.PERIPHERAL,
    ):
        """Constructor for the BluezLEAdvertisement"""
        # Call the super init with the type of interface
        super().__init__(name=BluezServiceInterfaces.ADVERTISEMENT_INTERFACE.value)

        # Advertisement values
        self.local_name = local_name
        self.ad_type = advertisement_type.value
        self.service_uuids = []
        self.manufacturer_data = {}
        self.service_data = {}

        # Note these are not available on every platform, see what the log of the supported capabilities are
        self.min_interval = 20
        self.max_interval = 20
        self.tx_power = 20
        self.appearance = 0x80  # default to generic computer

        # Dbus values
        self.application = None
        self.path = None

        # Wrapper class, this is the able advertisement that owns this advert
        self.parent = advert_wrapper

    def __str__(self):
        return f"BluezLEAdvertisement [{self.local_name}, {self.service_uuids}, {self.manufacturer_data}]"

    def __repr__(self):
        return str(self)

    def add_service_uuid(self, service_uuid: Union[str, uuid.UUID]) -> None:
        """
        Adds a service uuid to the advertisement object. Must follow the bluez spec for a service uuid
        otherwise bluez will reject the advertisement.

        :param service_uuid: What service uuid should be added to this advertisement.
        :type service_uuid: Union[str, uuid.UUID]
        :return: None
        """
        if str(service_uuid) not in self.service_uuids:
            self.service_uuids.append(str(service_uuid))

    def add_service_data(
        self, service_uuid: Union[str, uuid.UUID], service_data: Union[str, bytes]
    ) -> None:
        """
        Updates the service data of the advertisement to include the specified service data. Refer to the
        reference doc above for the limitations on service data.

        :raises ValueError: Will raise a value error if the service_uuid param does not correspond to a service
            uuid that this advertisement has.

        :param service_uuid: The uuid of the service to add the service data to, this must be a service uuid
            you have already added to the advertisement.
        :type service_uuid: Union[str, uuid.UUID]
        :param service_data: The service data to include in the advertisement under the service uuid.
        :type service_data: Union[str, bytes]
        :return: None
        """
        if str(service_uuid) not in self.service_uuids:
            raise ValueError(
                "Invalid service UUID provided, verify the service uuid was added to the advertisement"
            )

        if self.service_data.get(str(service_uuid)):
            logger.warning("Removing old service data from advertisement!")

        if isinstance(service_data, str):
            self.service_data[str(service_uuid)] = Variant(
                "ay", bytes(service_data.encode())
            )
        if isinstance(service_data, bytes):
            self.service_data[str(service_uuid)] = Variant("ay", service_data)

    async def update_service_data(
        self, service_uuid: Union[str, uuid.UUID], service_data: Union[str, bytes]
    ) -> None:
        """
        Update previously set service data with new data and refresh the advertisement using the bound application. This
        can be called after the `ABlePeripheralServer` that owns this advertisement has started advertising.

        :raises RuntimeError: Will raise an exception if this advertisement has not yet been added to an application,
            this is required for refreshing the service data.

        :param service_uuid: the uuid of the service whose service data you are updating.
        :type service_uuid: Union[str, uuid.UUID]
        :param service_data:  the service data to overwrite the previous service data with.
        :type service_data: Union[str, bytes]
        :return: None
        :rtype: None
        """
        if self.application is None:
            raise RuntimeError(
                "Tried to update service data without adding the advertisement to a running"
                "application, this will have no effect."
            )

        if service_uuid not in self.service_data:
            raise ValueError(
                "Invalid service UUID provided, there is no current service data"
            )

        if isinstance(service_data, str):
            self.service_data[service_uuid] = Variant(
                "ay", bytes(service_data.encode())
            )
        if isinstance(service_data, bytes):
            self.service_data[service_uuid] = Variant("ay", service_data)

        # Refresh the advertisement using the bound application
        await self.application.refresh_advertisement(advertisement=self.parent)

    def add_manufacturer_data(
        self, manufacturer_id: int, manufacturer_data: Union[str, bytes]
    ):
        """
        Updates the manufacturer data of the advertisement to include the specified manufacturer data. Refer to the
        reference doc above for the limitations on manufacturer data.

        TODO(Bernie, Ori): Add limitations for what the manufacturer data can be

        :param manufacturer_id: the id of the manufacturer to add the manufacturer data to, see the Bluetooth
            Special Interest groups website for what manufacturer id's are valid to be used in your use case.
        :type manufacturer_id: int
        :param manufacturer_data: the manufacturer data to add
        :type manufacturer_data: Union[str, bytes]
        :return: None
        :rtype: None
        """
        if self.manufacturer_data.get(manufacturer_id):
            logger.warning("Removing old manufacturer data from advertisement!")

        self.manufacturer_data[manufacturer_id] = Variant(
            "ay",
            bytes(
                manufacturer_data.encode()
                if isinstance(manufacturer_data, str)
                else manufacturer_data
            ),
        )

    @dbus_property(access=PropertyAccess.READ, name="Type")
    def _type(self) -> "s":  # type: ignore
        """
        Read only Dbus property to read the type of the advertisement.

        :return:  Determines the type of advertising packet requested. Possible values: "broadcast" or "peripheral"
        :rtype: str
        """
        return self.ad_type

    @dbus_property(access=PropertyAccess.READ, name="ServiceUUIDs")
    def _service_uuids(self) -> "as":  # type: ignore
        """
        Read only Dbus property to read the service uuids of the advertisement.

        :return:  List of UUIDs to include in the "Service UUID" field of
                        the Advertising Data.
        :rtype: list(str)
        """
        return self.service_uuids

    @dbus_property(access=PropertyAccess.READ, name="ServiceData")
    def _service_data(self) -> "a{sv}":  # type: ignore
        """
        Read only Dbus property to read the service data of the advertisement.

        :return: Service Data elements to include. The keys are the
                        UUID to associate with the data.
        :rtype: A dbus dictionary represnted by an array with strings as keys and variant entries which are
            arrays of bytes.
        """
        return self.service_data

    @dbus_property(access=PropertyAccess.READ, name="LocalName")
    def _local_name(self) -> "s":  # type: ignore
        """
        Read only Dbus property to local name of the advertisement.

        :return:  Local name to be used in the advertising report. If the
                        string is too big to fit into the packet it will be
                        truncated.
        :rtype: str
        """
        return self.local_name

    @dbus_property(access=PropertyAccess.READ, name="ManufacturerData")
    def _manufacturer_data(self) -> "a{qv}":  # type: ignore
        """
        Read only Dbus property to read the manufacturer data of the advertisement.

        :return: Manufacturer Data fields to include in the Advertising Data.  Keys are the Manufacturer ID
                        to associate with the data.
        :rtype: A dbus dictionary represented by an array with strings as keys and variant entries which are
            arrays of bytes.
        """
        return self.manufacturer_data

    @dbus_property(access=PropertyAccess.READ, name="MaxInterval")
    def _max_interval(self) -> "u":  # type: ignore
        """
        Dbus property which returns the max advertising interval if the advertising manager allows it.

        :return: The max interval in ms
        :rtype: uint32
        """
        return self.max_interval

    @dbus_property(access=PropertyAccess.READ, name="MinInterval")
    def _min_interval(self) -> "u":  # type: ignore
        """
        Dbus property which returns the min advertising interval if the advertising manager allows it.

        :return: The min interval in ms
        :rtype: uint32
        """
        return self.min_interval

    @dbus_property(access=PropertyAccess.READ, name="Appearance", disabled=True)
    def _appearance(self) -> "q":  # type: ignore
        """
        Appearance to be used in the advertising report. Possible values: as found on GAP Service.

        :return: The uint mapping to the appearance
        :rtype: uint16
        """
        return self.appearance

    @dbus_property(access=PropertyAccess.READ, name="TxPower", disabled=True)
    def _tx_power(self) -> "n":  # type: ignore
        """
        Dbus property which returns the transmission power of this advertising set if the advertising manager allows it.

        :return: The dBm this should be transmitted at [-127 + 20]
        :rtype: int16
        """
        return self.tx_power

    @method(name="Release")
    def _release(self):
        """
        This method gets called when the service daemon removes the Advertisement. A client can use it to do cleanup
        tasks. There is no need to call UnregisterAdvertisement because when this method gets called it has already
        been unregistered.

        TODO(Ori): Why does this not get called when we restart bluetoothd? See notion for this bug.

        :return: None
        :rtype: None
        """
        logger.debug(f"{self} has been released by bluez...")

        # Set the error state in the application
        self.application.error_state = True



#    Copyright (c)  2021  Allthenticate
"""
Abstraction of the ABleCharacteristic that all backend characteristics inherit from
"""

import abc
import asyncio
import typing
import uuid
from typing import TYPE_CHECKING, List, Optional, Union

if TYPE_CHECKING:
    from able import CharacteristicFlags
    from able.central import ABleCentral
    from able.service import ABleService


class ABleCharacteristic:
    """
    The abstract base class for all of the platform specific implementations of a BLE Characteristic. This class should
    not be used directly, similarly to the `ABleService`, the characteristic that will be imported will be specific
    to which platform ABle is being used on.

    :param characteristic_uuid: the uuid that this characteristic will have, best practice is that it is not the
        same as the service that will own it
    :param value: parameter to initialize the characteristic with a default value, defaults to b'0'
    :param flags: list of permissions/flags to initialize the characteristic with, defaults to read, write and notify
    :param nickname: provide a nickname to the characteristic that will show up in logs, defaults to None
    :param notify_on_write: if `True`, write from centrals will trigger a notification/indication to all subscribed
        centrals
    :ivar bytes value: the current value of the characteristic, setting this value on all backends will trigger an
        update to connected centrals.
    """

    __metaclass__ = abc.ABCMeta

    def __init__(
        self,
        characteristic_uuid: Union[str, uuid.UUID],
        value: bytes = bytes(0),
        flags: List = None,
        nickname: str = None,
        notify_on_write: bool = False,
        **kwargs,
    ):
        """Initialize the BLECharacteristic"""
        from able import CharacteristicFlags  # pylint: disable=import-outside-toplevel

        default_flags = [
            CharacteristicFlags.READ,
            CharacteristicFlags.WRITE,
            CharacteristicFlags.NOTIFY,
        ]
        # Handle the flags
        if flags is None:
            flags = default_flags

        # Handle members shared across all platforms
        self._flags = flags
        self.uuid: str = str(characteristic_uuid)
        self.nickname = nickname
        self.notify_on_write = notify_on_write
        self.client_subscribed: asyncio.Event = asyncio.Event()

        self.service: Optional["ABleService"] = None

        # Work as a mixin
        super().__init__(**kwargs)  # type: ignore

        # Goes last since some __setattr__'s are overloaded
        self.value = value

    def __key(self):
        return str(self.service.uuid), str(self.uuid)

    def __hash__(self):
        return hash(self.__key())

    @abc.abstractmethod
    async def notify_central_of_change(
        self,
        central: "ABleCentral",
        data: bytes,
    ) -> None:
        """
        Virtual method for notifying a central of a change on a characteristic, this is how we handle sending
        data to a central under the hood. The characteristic sets its value to the data then notifies the central
        of the change.

        :param central: the central to notify of the change
        :param data: the data to send to the central by setting the value of the characteristic then notifying
            or indicating only that central of the change
        :return: None
        """
        raise NotImplementedError("This is a purely virtual method...")

    async def wait_for_client_subscription(self, timeout=5) -> None:
        """
        Helper function that will block until a client subscribes to this characteristic. Note that in the current
        implementation this can only check if any client subscribed rather than blocking til as specific central
        subscribes, this is actively being worked on.

        TODO(Bernie): Client specific blocking

        :param timeout: How long to wait before timing out
        """
        await asyncio.wait_for(self.client_subscribed.wait(), timeout=timeout)

    @property
    def flags(self):
        """Get the value of the char flags as a list"""
        return [flag.value for flag in self._flags]

    @property
    def notifying(self):
        """Notifying property checks to see if the client subscribed event has been set"""
        return self.client_subscribed.is_set()

    async def broadcast(self, data: typing.Union[bytes, str]):
        """Virtual method for broadcasting, needs to be implemented on a platform specific basis"""
        raise NotImplementedError



#    Copyright (c)  2021  Allthenticate
"""
Home to any utility classes and functions in ABle
"""


class DeviceCallbackData:
    """
    Akin to the AdvertisementData class that Bleak implements, this call will be used for the standardizing
    of all callback data between the bless interface and anything that utilizes it.
    """

    def __init__(self, devices_identifier: str, **kwargs):
        """
        Initializes the DeviceCallbackData object
        :param devices_identifier: required param with the unique identifier of the ble device. It could be
        an MAC address (eg. 52:B8:E9:9E:F5:E5), UUID and etc. as long as it's unique.
        :type devices_identifier: `str`
        :param args: Extra arguments
        :type args:
        :param kwargs: Extra arguments
        :type kwargs:
        """
        # Update the identifier in the correct format
        self.identifier = devices_identifier.replace("_", ":")

        # Characteristic uuid in case of subscription/unsubscription
        self.char_uuid = kwargs.get("char_uuid", None)

        # Extra data TODO
        self.device_data = kwargs.get("device_data", {})

    def __str__(self):
        return f"DeviceCallbackData for: {self.identifier}"

    def __repr__(self):
        return str(self)



#    Copyright (c)  2021  Allthenticate
"""
For reference of the implementation of this Gatt service abstraction, see:
https://gitlab.com/allthenticate/dependencies/bluez/-/blob/master/doc/gatt-api.txt
https://dbus.freedesktop.org/doc/dbus-tutorial.html
"""

# Native libraries
import logging
import uuid
from typing import TYPE_CHECKING, Dict, Optional

# Dependencies
from dbus_next.constants import PropertyAccess  # type: ignore
from dbus_next.service import ServiceInterface, dbus_property  # type: ignore

# Able Dependencies
from able.bluezdbus import BluezServiceInterfaces

# Setup logging
from able.service import ABleService

logger = logging.getLogger(name=__name__)

if TYPE_CHECKING:
    from able.characteristic import ABleCharacteristic


class BluezService(ABleService, ServiceInterface):
    """
    Class implementation of a BLE GATT Service using Dbus-Next. This class inherits from the `ABleService` so that
    it shares the same params for its constructor and supports the same capabilities as the services on other
    platforms. It also inherits from `ServiceInterface` so that this class can be exported on the DBus so that
    Bluez is able to get a proxy for this with the "org.bluez.GattService1" interface.

    :param service_uuid: the uuid for this service
    :param is_primary: whether or not this is a primary service, secondary services are rarely used as they
        are intended to be used in other services, defaults to True
    :ivar characteristics: a dictionary of uuid string - characteristic pairs, this is used to fetch a characteristic
        from a service if a reference to the characteristic is lost in your program.
    """

    def __init__(self, service_uuid: uuid.UUID, is_primary: bool = True):
        """Initialize the BluezService"""
        super().__init__(
            service_uuid, name=BluezServiceInterfaces.GATT_SERVICE_INTERFACE.value
        )

        # DBus values
        self.primary = is_primary
        self._path: Optional[str] = None

        # Additional members
        self.characteristics: Dict[str, "ABleCharacteristic"] = {}

    def __str__(self):
        return f"BluezLEService [{self.uuid}]"

    @dbus_property(access=PropertyAccess.READ, name="UUID")
    def _uuid(self) -> "s":  # type: ignore
        """
        Read only property for the bluez service to get the uuid of this gatt service interface.

        :return: 128-bit service UUID.
        """
        return str(self.uuid)

    @dbus_property(access=PropertyAccess.READ, name="Primary")
    def _primary(self) -> "b":  # type: ignore
        """
        Read only property for the bluez service to get whether or not this service is the primary service
        for the application.

        :return: Indicates whether or not this GATT service is a primary service. If false, the service is secondary.
        """
        return self.primary

    @property
    def path(self) -> str:
        """Path property, raises an exception if not properly defined"""
        if self._path is None:
            raise RuntimeError(
                "Attempting to fetch path of an unset service, has it been added correctly?"
            )
        return self._path



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['able', 'able.bluezdbus', 'able.corebluetooth', 'able.plugins']

package_data = \
{'': ['*']}

extras_require = \
{':sys_platform == "darwin"': ['pyobjc>=7.3,<8.0'],
 ':sys_platform == "linux"': ['dbus-next>=0.2.2,<0.3.0']}

setup_kwargs = {
    'name': 'able',
    'version': '0.3.4',
    'description': "ABle (Allthenticate's Bluetooth Low Energy (Library)) is a platform agnostic python framework for communication with centrals as a BLE Peripheral with simple yet powerful 'BLE Sockets'",
    'long_description': '# ABle\n[![Imports: isort](https://img.shields.io/badge/%20imports-isort-%231674b1?style=flat&labelColor=ef8336)](https://pycqa.github.io/isort/)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Type checking: mypy](https://camo.githubusercontent.com/59eab954a267c6e9ff1d80e8055de43a0ad771f5e1f3779aef99d111f20bee40/687474703a2f2f7777772e6d7970792d6c616e672e6f72672f7374617469632f6d7970795f62616467652e737667)](https://mypy.readthedocs.io/en/stable/)\n\nAble stands for Allthenticate\'s BLE Peripheral Library.\nIt serves the purpose of enabling the abstraction of using a BLE Peripheral on Ubuntu, MacOs and Windows based systems\nwithout having to adapt your software to have platform specific support.\n\n# In Progress\n\nNote that this project is still undergoing work by the development team, the main things we are working on are:\n\n- Cleaning up some bugs in the bluezdbus and macos backend\n- Making the gitlab/github repo public for contributions\n- Setup a slack for developers to reach out to us about issues or ideas\n- Make logging/use clearer and update our documentation (for example we require a fork of bluez) and moving things from this readme to our docs\n- Get our docs on read the docs\n- Some more surprises :)\n\n# How To\'s\n\n## Quick Start\nTo get started just run `get_started.sh`. This will install poetry and all of the project\'s dependencies, and drop you into the projects virtual environment.\n\n```\nbash get_started.sh\n```\nWhenever you pull new updates we recommend running a quick `poetry install` to get any updates. From there, please check out our examples to get started with the project!\n\n## Upgrading Poetry Installer\n\nIt may be the case that the current version of your Poetry package is outdated. Run `scripts/check_poetry.sh` to check if there is a newer version of the poetry installer.\n\n```\nbash scripts/check_poetry.sh\n```\n\n# Documentation\n\n## Writing Docs\nWe follow the Sphinx docstring format for our docstrings, see the following [site](https://sphinx-rtd-tutorial.readthedocs.io/en/latest/docstrings.html) on the complete specification, but the\ngeneral docstring will look like:\n\n```\n"""\n[Summary]\n\n:param [ParamName]: [ParamDescription], defaults to [DefaultParamVal]\n:type [ParamName]: [ParamType](, optional)\n...\n:raises [ErrorType]: [ErrorDescription]\n...\n:return: [ReturnDescription]\n:rtype: [ReturnType]\n"""\n```\n\n## Generating Docs\nTo generate the docs for local use you should just have to run the following inside of the poetry shell:\n\n```\nmake html\n```\nThen open the html produced by Sphinx.\n\n## Auto-Generating Docs\n\nIf you want to have docs magically update as you are writing them, run the following command:\n\n```\nsphinx-autobuild source/ build/html/\n```\n\n# Testing Able\n\n## How To Test?\nRunning tests are easy, you need to only do a few things as a developer. First track down the IP of the companion you\nplan to be using, we recommend a raspberry pi. Once you have that IP, export it as an environment variable as so:\n\n```\nexport ABLE_CENTRAL_IP="<IP>"\n```\n\nNow, we can run the tests with the following command:\n\n```\npoetry install && poetry run pytest tests\n```\n\n## Rigorous Testing\nIf you want to do more rigorous testing over a long period of time and check for flaky tests, you will\nhave to modify the `pyproject.toml`. We already have the dependencies you need to run tests multiple times\nto detect flakiness, all you need to do is modify the following line:\n\n```\naddopts = "--flake-finder --flake-runs=1 --reruns 5 --reruns-delay 5"\n```\nInto:\n```\naddopts = "--flake-finder --flake-runs=10 --reruns 5 --reruns-delay 5"\n```\nThis will run each test 10 times, you can even modify it to be greater should you choose. You can also\nmodify the `reruns` and `reruns-delay` parameters to change how much time you should wait between failed tests,\nmaybe to let things simmer and how many reruns you will accept.\n\n### Speedy Tests\nOur dependencies include `pytest-fast-first` which will locally track which tests are quicker and will use AI and\ndeep learning (a json dictionary of times) to track and run tests that go faster first! Neat!\n\n## Coming Soon To Testing\nWe are hoping to have unit tests coming soon for Able but right now are relying solely on hardware in the loop tests\nto get things off the ground. Eventually we will detect if you have a companion set and if not, we will only run\nthe unit tests.\n\n# Support\nIf you have any questions on the use of this library feel free to reach out the head\nmaintainer <bernie@allthenticate.com> or submit an issue on the repository.\n\n## Contributing\nContributing is not currently enabled but once the repository is licensed we will be opening the project up for public contributions.\n\n\n# Acknowledgements\n\nThis project was inspired by the great work done by the developer team for [Bleak](https://github.com/hbldh/bleak)\nwhich is a fantastic platform agnostic bluetooth framework for a BLE client/central we would highly reccomend!\n\nWe also took notes from the work done by Kevin Car with his companion library [Bless](https://github.com/kevincar/bless),\nwho made a great server supplement to Bleak whose work saved us from countless hours from fighting dbus and pyobjc!\n',
    'author': 'Bernie Conrad',
    'author_email': 'bernie@allthenticate.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://allthenticate.com/',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.6.2,<4.0',
}


setup(**setup_kwargs)
